/*
                ***** alloc2D.c *****

Routines to dynamically allocate and free 2-D arrays in C.

Author: John M. Weiss, Ph.D.
Written 10/02/86 (Ridge 32C UNIX system).

Modifications:
05/20/91 - Rewrite as single function, use #defines to get all data types.
10/14/91 - Turbo C port. Use far pointers explicitly so that 2-D memory
    allocation is not crippled by the small memory model. Use huge
    pointer for setting up array so that it can be allocated as one
    large buffer. Note that memory allocation appears to be handled
    like a LIFO stack. Deallocated "holes" in memory are not necessarily
    available until all the heap space above becomes free.
06/20/93 - Rewrite for both Turbo C and DJGPP.
*/

#include <stdio.h>

#ifdef __TURBOC__
#include <alloc.h>
#else
#include <malloc.h>
#define far
#define huge
#define farcalloc(n,s)  calloc(n,s)
#define farfree(ptr)    free(ptr)
#endif

/* Allocate storage for 2-D array of any type: */
void far **alloc2D (int nrows, int ncols, int size)
{
    void far **array;
    char huge *buffer;
    int row;

    if ((array = (void far **)calloc(nrows, sizeof(void far *))) == NULL)
        return NULL;

    if ((buffer = (char huge *)farcalloc(nrows * (long)ncols, size)) == NULL)
    {
        free(array);
        return NULL;
    }

    for (row = 0; row < nrows; row++, buffer += ncols * (long)size)
	array[row] = (void far *)buffer;

    return array;
}

/* Free a dynamically allocated 2-D arrays: */
void free2D (void far **array)
{
    farfree(*array);
    free(array);
}

