/* ---------------------------------------------------------------------- */
/*                        FULL SCREEN DEBUGGER                            */
/*                                                                        */
/* Copyright 1994 by Morten Welinder, terra@diku.dk                       */
/* ---------------------------------------------------------------------- */
/* 

This debugger is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This debugger is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with djgpp; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */
/* ---------------------------------------------------------------------- */
#define FULLSCR_VERSION 0.90

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <pc.h>
#include <keys.h>
#define K_BackTab 0x10f
#define K_ELeft 0x24b
#include <setjmp.h>
#define far
#include "ed.h"
#define A_black 0
#include "syms.h"
#include "paging.h"
#include "unassmbl.h"
/* ---------------------------------------------------------------------- */
typedef struct {
  word16 sig0;
  word16 sig1;
  word16 sig2;
  word16 sig3;
  word16 exponent:15;
  word16 sign:1;
} NPXREG;

typedef struct {
  word32 control;
  word32 status;
  word32 tag;
  word32 eip;
  word32 cs;
  word32 dataptr;
  word32 datasel;
  NPXREG reg[8];
} NPX;
/* ---------------------------------------------------------------------- */
int can_longjmp = 0;
jmp_buf debugger_jmpbuf;

static int debug_screen_p;
static char *user_screen_save, *debug_screen_save;
static int screen_attr;
static int screen_attr_normal;
static int screen_attr_focus;
static int screen_attr_break;
static int screen_attr_message;
static int cols;
static int rows;

static word32 data_dump_origin, data_dump_last;
static word32 code_dump_origin, code_dump_last;
static word32 code_pane_pos[50];
static int code_pane_active;
static int npx_pane_active;

#define MAXINSTLEN 16
#define PANECOUNT 6

static int pane;
static int pane_positions[PANECOUNT], pane_pos;
static int first_step;
static word32 main_entry;
char read_buffer[200];
static NPX npx;
/* ---------------------------------------------------------------------- */
static void message (int, char *, ...);
static char *get_screen (void);
static void put_screen (char *);
static void debug_screen (void);
static void user_screen (void);
static void redraw (int);
static int read_eval (int *, char *);
/* ---------------------------------------------------------------------- */
#define fs_put(x,y,txt) \
  { short unsigned *p = (short unsigned *)(screen + 3 + (80 * (y) + (x)) * 2);\
    unsigned char *q = (unsigned char *)(txt); \
    while (*q) (*p++ = screen_attr + *q++); }

#define fs_draw(x,y,ch,delta,count) \
  { short unsigned *p = (short unsigned *)(screen + 3 + (80 * (y) + (x)) * 2),\
                   attrch = screen_attr + (unsigned char) (ch);\
    int i = count;\
    while (i--) { *p = attrch; p += delta; }; }
/* ---------------------------------------------------------------------- */
static
void save_npx (void)
{
asm(                                                                   "\n\
        inb     $0xa0,%al                                               \n\
        testb   $0x20,%al                                               \n\
        jz      Lfclex_done                                             \n\
        xorl    %eax,%eax                                               \n\
        outb    %al,$0xf0                                               \n\
        movb    $0x20,%al                                               \n\
        outb    %al,$0xa0                                               \n\
        outb    %al,$0x20                                               \n\
Lfclex_done:                                                            \n\
	movl	$_npx, %eax						\n\
	fnsave	(%eax)							\n\
	fwait								"
    );
}
/* ---------------------------------------------------------------------- */
static void
load_npx (void)
{
asm(                                                                   "\n\
        movl    $_npx, %eax                                             \n\
        movb    $0,4(%eax)              /* clear pending exceptions */  \n\
        frstor  (%eax)                                                  "
    );
}
/* ---------------------------------------------------------------------- */
inline static void
highlight (char *screen, int x, int y, int len)
{
  short unsigned *p = (short unsigned *)(screen + 4 + (80 * y + x) * 2);
  char attr = screen_attr >> 8;
  while (len--) 
    { *(char *)p = attr; p++; };
}
/* ---------------------------------------------------------------------- */
static void
frame (char *screen, int x1, int y1, int x2, int y2)
{
  fs_draw (x1 + 1, y1, '', 1,    x2 - x1 - 1);
  fs_draw (x1 + 1, y2, '', 1,    x2 - x1 - 1);
  fs_draw (x1, y1 + 1, '', cols, y2 - y1 - 1);
  fs_draw (x2, y1 + 1, '', cols, y2 - y1 - 1);
  fs_put (x1, y1, "");
  fs_put (x2, y1, "");
  fs_put (x1, y2, "");
  fs_put (x2, y2, "");
}
/* ---------------------------------------------------------------------- */
static void
message (int class, char *fmt, ...)
{
  char buf[80], *screen, *save;
  int len, saveattr = screen_attr;

  vsprintf (buf, fmt, (&fmt) + 1);
  len = strlen (buf);
  save = get_screen ();
  screen = get_screen ();
  screen_attr = screen_attr_message << 8;
  frame (screen, 1, rows / 2 - 1, cols - 2, rows / 2 + 1);
  fs_draw (2, rows / 2, ' ', 1, cols - 4);
  fs_put ((cols - len) / 2, rows / 2, buf);
  screen_attr = saveattr;
  put_screen (screen);
  switch (class)
    {
    case 0:
      sleep (1);
      break;
    case 1:
      (void) getxkey ();
    }
  put_screen (save);
  free (save);
  free (screen);
}
/* ---------------------------------------------------------------------- */
static int
read_string (char *starttext)
{
  char *screen, *save;
  int key, esc, pos, leave = 0, y = rows / 2;

  save = get_screen ();
  screen = get_screen ();
  strcpy (read_buffer, starttext);
  pos = strlen (read_buffer);

  frame (screen, 1, y - 1, cols - 2, y + 1);
  do
    {
      fs_draw (2, y, ' ', 1, cols - 4);
      fs_put (3, y, read_buffer);
      pos = strlen (read_buffer);
      put_screen (screen);
      ScreenSetCursor (y, 3 + pos);
      key = getxkey ();
      switch (key)
        {
        case K_Return:
          leave = 1;
          esc = 0;
          break;
        case K_Escape:
          leave = 1;
          esc = 1;
	  read_buffer[0] = '\0';
          break;
        case K_BackSpace:
          if (pos > 0)
            read_buffer[--pos] = '\0';
          break;
        default:
          if (key >= ' ' && key <= 0xff && pos <= cols - 7)
            {
              read_buffer[pos++] = key & 0xff;
              read_buffer[pos] = '\0';
            }
        }
    }
  while (!leave);
  free (screen);
  put_screen (save);
  free (save);
  return esc;
}
/* ---------------------------------------------------------------------- */
static int
eval (char *expr, int *okp)
{
  word32 sym;

  sym = syms_name2val (read_buffer);
  if (undefined_symbol)
    {
      *okp = 0;
      message (1, "Expression not understood");
      return 0;
    }
  else
    {
      *okp = 1;
      return (int) sym;
    }
}
/* ---------------------------------------------------------------------- */
static int
read_eval (int *okp, char *starttext)
{
  *okp = !read_string (starttext);
  if (*okp && read_buffer[0] != '\0')
    return eval (read_buffer, okp);
  else
    return 0;
}
/* ---------------------------------------------------------------------- */
static int
valid_addr (word32 vaddr, int len)
{
  int a;

  len--;
  for (a = 0; a < MAX_AREA; a++)
    if ((vaddr + len <= areas[a].last_addr) && (vaddr >= areas[a].first_addr))
      return 1;
  return 0;
}
/* ---------------------------------------------------------------------- */
inline static int
valid_instaddr (word32 vaddr)
{
  return valid_addr (vaddr, MAXINSTLEN);
}
/* ---------------------------------------------------------------------- */
static int
get_breakpoint (word32 v)
{
  int b;

  for (b = 0; b < 4; b++)
    if ((edi.dr[7] & (3 << (b * 2))) != 0 &&
        (edi.dr[b] == v + edi.app_base))
      return b;
  return -1;
}
/* ---------------------------------------------------------------------- */
static int
reset_breakpoint (int no)
{
  edi.dr[7] &= ~(2 << (no * 2));
}
/* ---------------------------------------------------------------------- */
static int
set_breakpoint (int no, word32 v)
{
  int b;

  for (b = 0; b < 4; b++)
    if ((b == no) || ((b > 0) && (edi.dr[7] & (3 << (b * 2))) == 0))
      {
        edi.dr[7] |= 2 << (b * 2);
        edi.dr[b] = v + edi.app_base;
        return b;
      }
  return -1;
}
/* ---------------------------------------------------------------------- */
static int
code_skip (int origin, int count)
{
  int len, *next, i, j, k, instcount, done, leave;
  char *state, *inst, *source;

  if (count >= 0)
    {
      while (count-- > 0)
	if (valid_instaddr (origin))
	  {
	    (void) unassemble_proper (origin, &len);
	    origin += len;
	  }
	else
	  origin ++;
      return origin;
    }
  else
    {
      count = -count;
      instcount = MAXINSTLEN * (count + 16) + 1;
      next = alloca (instcount * sizeof (int));
      memset (next, 0, instcount * sizeof (int));
      state = alloca (instcount * sizeof (char));
      memset (state, 0, instcount * sizeof (char));

      done = 0;
      do
	{
	  for (i = 0; i < 2 * count; i++)
	    {
	      done++;
	      j = origin - done;
	      if (valid_instaddr (j))
		{
		  inst = unassemble_proper (j, &len);
		  source = unassemble_source (j);
		  next[done] = j + len;
		  if (source)
		    {
		      leave = 0;
		      k = done;
		      while (j < origin && state[k] == 0 && !leave)
			{
			  state[j++,k--] = 2;
			  while (len-- > 1)
			    state[j++,k--] = 3;
			  leave = (strncmp (inst, "jmp", 3) == 0
				   || strncmp (inst, "ret", 3) == 0
				   || strncmp (inst, "iret", 4) == 0);
			  if (!leave)
			    inst = unassemble_proper (j, &len);
			}
		    }
		}
	      else
		{
		  state[done] = 2;
		  next[done] = j + 1;
		}
	    }
	  j = 1;
	  k = count;
	  while (k > 0 && j <= done && state[j] >= 2)
	    if (state[j++] == 2)
	      k--;
	}
      while (k > 0 && done + 2 * count <= instcount);
      if (k == 0)
	return origin - j + 1;
      else
	{
	  i = origin;
	  k = 0;
	  while (count > 0 && k <= done)
	    {
	      leave = 0;
	      j = MAXINSTLEN;
	      while (!leave && j > 1)
		if (j + k <= done && next[j + k] == i)
		  leave = 1;
		else
		  j--;
	      if (!leave)
		i--, k++;
	      else
		i -= j, k += j;
	      count--;
	    }
	  return i;
	}
    }
}
/* ---------------------------------------------------------------------- */
static void
code_pane_goto (word32 v)
{
  int i = 0;

  if (v >= code_dump_origin && v <= code_dump_last)
    {
      while (code_pane_pos[i] < v)
        i++;
      if (code_pane_pos [i] == code_pane_pos[i+1])
        i++;
    }
  else
    {
      code_dump_origin = v;
      if (valid_instaddr (v) && unassemble_source (v) != NULL)
        i++;
    }
  if (pane == 0)
    pane_pos = i;
  else
    pane_positions[0] = i;
}
/* ---------------------------------------------------------------------- */
static void
step (int kind)
{
  int i, len, olddr7, terminated, refmem;
  char *inst;

  switch (kind)
    {
    case 0:
      inst = unassemble_proper (a_tss.tss_eip, &len);
      if (strcmp (inst, "popf") == 0 || strcmp (inst, "pushf") == 0)
        kind = 1;
      break;
    case 1:
      if (first_step)
        kind = 3;
      else
        {
          inst = unassemble_proper (a_tss.tss_eip, &len);
          if (strncmp (inst, "loop", 4)
              && strncmp (inst, "call", 4)
              && strncmp (inst, "int", 3) )
            kind = 0;
          break;
        }
    }
  switch (kind)
    {
    case 0: /* "Trace" */
      a_tss.tss_eflags |= 0x0100;
      olddr7 = edi.dr[7];
      edi.dr[7] = 0;
      refmem = strchr (inst, '[') != 0;
      if (refmem)
        user_screen ();
      run_child ();
      edi.dr[7] = olddr7;
      a_tss.tss_eflags &= ~0x0100;
      break;
    case 1: /* "Step Over" */
      user_screen ();
      set_breakpoint (0, a_tss.tss_eip + len);
      run_child ();
      reset_breakpoint (0);
      break;
    case 2: /* "Run" */
      user_screen ();
      run_child ();
      break;
    case 3: /* "Run to `_main'" */
      set_breakpoint (0, main_entry);
      user_screen ();
      run_child ();
      reset_breakpoint (0);
    }
  first_step = 0;
  i = a_tss.tss_irqn;
  terminated = (i == 0x21) && (a_tss.tss_eax & 0xff00) == 0x4c00;
  if (terminated)
    a_tss.tss_eip -= 2;  /* point to int 21h */
  code_pane_goto (a_tss.tss_eip);
  debug_screen ();
  redraw (0);
  if (terminated)
    message (1, "Program terminated normally, exit code is %d",
             (word8)a_tss.tss_eax);
  else
    {
      int b;

      if (i != 1)
        {
          if (i == 0x79)
            message (0, "Keyboard interrupt");
          else if (i == 0x75)
            {
              char *reason;

              save_npx ();
              if ((npx.status & 0x0241) == 0x0241)
                reason = "stack overflow";
              else if ((npx.status & 0x0241) == 0x0041)
                reason = "stack underflow";
              else if (npx.status & 1)
                reason = "invalid operation";
              else if (npx.status & 2)
                reason = "denormal operand";
              else if (npx.status & 4)
                reason = "divide by zero";
              else if (npx.status & 8)
                reason = "overflow";
              else if (npx.status & 16)
                reason = "underflow";
              else if (npx.status & 32)
                reason = "loss of precision";
              else
                reason = "?";
              message (1, "Numeric Exception (%s) at eip=0x%08lx",
                       reason, npx.eip);
              load_npx ();
            }
          else
            if (i == 8 || (i >= 10 && i <= 14))
              message (1, "Exception %d (%#02x) occurred, error code=%#lx",
                       i, i, a_tss.tss_error);
            else
              message (1, "Exception %d (%#02x) occurred", i, i);
        }
      for (b = 0; b < 3; b++)
        if ((edi.dr[6] & (1 << b)) && (edi.dr[7] & (3 << (b * 2))))
          message (0, "Breakpoint %d hit", b);
    }
}
/* ---------------------------------------------------------------------- */
static void
put_screen (char *screen)
{
  switch (*screen++)
    {
    case 0:
      ScreenClear ();
      break;
    case 1:
      ScreenSetCursor (screen[0], screen[1]);
      ScreenUpdate (screen + 2);
      break;
    }
}
/* ---------------------------------------------------------------------- */
static char *
get_screen ()
{
  char *p;
  int r, c;

  p = malloc (cols * rows * 2 + 3);
  ScreenGetCursor (&r, &c);
  p[0] = 1;
  p[1] = r;
  p[2] = c;
  ScreenRetrieve (p + 3);
  return p;
}
/* ---------------------------------------------------------------------- */
void
debug_screen ()
{
  if (!debug_screen_p)
    {
      user_screen_save = get_screen ();
      put_screen (debug_screen_save);
      free (debug_screen_save);
      debug_screen_p = 1;
    }
}
/* ---------------------------------------------------------------------- */
void
user_screen ()
{
  if (debug_screen_p)
    {
      debug_screen_save = get_screen ();
      put_screen (user_screen_save);
      free (user_screen_save);
      debug_screen_p = 0;
    }
}
/* ---------------------------------------------------------------------- */
static void
redraw (int first)
{
  char buf[80];
  char *screen;
  char flags[] = "c?p?a?zn?ido????";

  debug_screen ();

  ScreenAttrib = screen_attr_normal;
  screen_attr = ScreenAttrib << 8;

  if (first) {
    ScreenClear ();
    ScreenSetCursor (0, 0);
  }
  screen = get_screen ();

  if (first)
    {
      frame (screen, 0, 0, cols - 1, rows - 1);
      frame (screen, cols - 5, 0, cols - 1, 17);
      frame (screen, cols - 18, 0, cols - 5, 17);
      frame (screen, cols - 33, 17, cols - 1, rows - 1);
      frame (screen, 0, 17, cols - 33, rows - 1);

      fs_put (cols - 5, 0, "");
      fs_put (cols - 5, 17, "");
      fs_put (cols - 1, 17, "");
      fs_put (0, 17, "");
      fs_put (cols - 18, 17, "");
      fs_put (cols - 33, 17, "");
      fs_put (cols - 18, 0, "");
      fs_put (cols - 33, rows - 1, "");
    }

  /* Show register status */
  sprintf (buf, "eax %08lx",     a_tss.tss_eax);  fs_put (cols - 17,  1, buf);
  sprintf (buf, "ebx %08lx",     a_tss.tss_ebx);  fs_put (cols - 17,  2, buf);
  sprintf (buf, "ecx %08lx",     a_tss.tss_ecx);  fs_put (cols - 17,  3, buf);
  sprintf (buf, "edx %08lx",     a_tss.tss_edx);  fs_put (cols - 17,  4, buf);
  sprintf (buf, "esi %08lx",     a_tss.tss_esi);  fs_put (cols - 17,  5, buf);
  sprintf (buf, "edi %08lx",     a_tss.tss_edi);  fs_put (cols - 17,  6, buf);
  sprintf (buf, "ebp %08lx",     a_tss.tss_ebp);  fs_put (cols - 17,  7, buf);
  sprintf (buf, "esp %08lx",     a_tss.tss_esp);  fs_put (cols - 17,  8, buf);
  sprintf (buf, " cs %04x    ",  a_tss.tss_cs);   fs_put (cols - 17,  9, buf);
  sprintf (buf, " ds %04x    ",  a_tss.tss_ds);   fs_put (cols - 17, 10, buf);
  sprintf (buf, " es %04x    ",  a_tss.tss_es);   fs_put (cols - 17, 11, buf);
  sprintf (buf, " fs %04x    ",  a_tss.tss_fs);   fs_put (cols - 17, 12, buf);
  sprintf (buf, " gs %04x    ",  a_tss.tss_gs);   fs_put (cols - 17, 13, buf);
  sprintf (buf, " ss %04x    ",  a_tss.tss_ss);   fs_put (cols - 17, 14, buf);
  sprintf (buf, "eip %08lx",     a_tss.tss_eip);  fs_put (cols - 17, 15, buf);
  sprintf (buf, "flg %08lx", a_tss.tss_eflags);   fs_put (cols - 17, 16, buf);

  {
    /* Show flags status */
    int f, y = 1;
    for (f = 0; flags[f]; f++)
      if (flags[f] != '?')
        {
          sprintf (buf, "%c=%d", flags[f], (int)((a_tss.tss_eflags >> f) & 1));
          fs_put (76, y++, buf);
        }
  }

  {
    /* Show breakpoints */
    int b, i, y = 18;
    long a;
    char *name;
    int32 delta;

    sprintf (buf, "%-31s", "");
    for (i = y; i < rows - 1; i++)
      fs_put (cols - 32, i, buf);

    for (b = 1; b < 4; b++, y += 2)
      if ((edi.dr[7] & (3 << (b * 2))) != 0)
        {
          a = edi.dr[b] - edi.app_base;
          sprintf (buf, "Breakpoint %d at %08lx", b, a);
          fs_put (cols - 32, y, buf);
          name = syms_val2name (a, &delta);
          if (name[0] != '0')
            {
              if (delta)
                sprintf (buf, "%s+%#lx", name, delta);
              else
                sprintf (buf, "%s", name);
              buf[30] = '\0';
              fs_put (cols - 31, y + 1, buf);
            }
        }
      else
        {
          sprintf (buf, "Breakpoint %d is not set", b);
          fs_put (cols - 32, y, buf);
        }
  }

  {
    /* Show data dump */
    word32 p = data_dump_origin;
    int x, y = 18, ok;
    unsigned char data[8];

    while (y < rows - 1)
      {
        if (ok = valid_addr (p, 8))
          read_child (p, data, (unsigned) 8);
        else
          for (x = 0; x < 8; x++)
            if (valid_addr (p + x, 1))
              read_child (p + x, data + x, (unsigned) 1);
        sprintf (buf, "%08lx: %02x %02x %02x %02x %02x %02x %02x %02x    ",
                 p,
                 data[0], data[1], data[2], data[3],
                 data[4], data[5], data[6], data[7]);
        buf[45] = '\0';
        for (x = 0; x < 8; x++)
          if (ok || valid_addr (p + x, 1))
            buf[37 + x] = data[x] ? data[x] : ' ';
          else
            {
              buf[37 + x] = ' ';
              buf[10 + x * 3] = '?';
              buf[11 + x * 3] = '?';
            }
        fs_put (1, y, buf);
        p += 8;
        y++;
      }
    data_dump_last = p - 1;
  }

  if (code_pane_active)
  {
    /* Show code dump */
    word32 p = code_dump_origin;
    int y = 1, len, source = 0;
    char *txt;

    while (y < 17)
      {
        source = !source;
        code_pane_pos[y - 1] = p;
        if (source)
          {
            txt = unassemble_source (p);
            if (txt)
              {
                sprintf (buf, "%*s", -(cols - 19), txt);
                fs_put (1, y, buf);
                y++;
              }
          }
        else
          {
            if (valid_instaddr (p))
              txt = unassemble_proper (p, &len);
            else
              txt = "?", len = 1;
            sprintf (buf, "%08lx%c%*s",
                     p,
                     p == a_tss.tss_eip ? 16 : ' ',
                     -(cols - 28), txt);
            screen_attr =
              (get_breakpoint (p) == -1
               ? screen_attr_normal : screen_attr_break) << 8;
            fs_put (1, y, buf);
            screen_attr = screen_attr_normal << 8;
            p += len;
            y++;
          }
      }
    code_pane_pos[y - 1] = p + 1;
    code_dump_last = p - 1;
  }

  if (npx_pane_active)
  {
    /* Show npx dump */
    int i, y = 1;

    save_npx ();
    sprintf (buf, "Control: 0x%04lx%*s", 
	     npx.control & 0xffff, -(cols - 34), "");
    fs_put (1, y++, buf);
    sprintf (buf, "Status:  0x%04lx%*s", 
	     npx.status & 0xffff,  -(cols - 34), "");
    fs_put (1, y++, buf);
    sprintf (buf, "Tag:     0x%04lx%*s", 
	     npx.tag & 0xffff,  -(cols - 34), "");
    fs_put (1, y++, buf);

    for (i = 0; i < 8; i++)
      {
	double d;
	int tag;
	int tos = (npx.status >> 11) & 7;
	char *dstr;

	tag = (npx.tag >> (((i + tos) % 8) * 2)) & 3;
	switch (tag)
          {
	  case 0:
	    if (((int)npx.reg[i].exponent-16382 < 1000) &&
		((int)npx.reg[i].exponent-16382 > -1000))
              {
                d = (((npx.reg[i].sig0 / 65536.0
		       + npx.reg[i].sig1) / 65536.0
		      + npx.reg[i].sig2) / 65536.0
		     + npx.reg[i].sig3) / 65536.0;
                d = ldexp (d, (int) npx.reg[i].exponent - 16382);
                if (npx.reg[i].sign)
                  d = -d;
		dstr = alloca (30);
                sprintf(dstr,"%+.16g", (double) d);
              }
	    else
	      dstr = "Valid, but huge";
	    break;
	  case 1:
	    dstr = (npx.reg[i].sign) ? "-Zero" : "+Zero";
	    break;
	  case 2:
	    dstr = (npx.reg[i].sign) ? "-Special" : "+Special";
	    break;
	  case 3:
	    dstr = "Empty";
	    break;
          }
	sprintf (buf, "st(%d): %d %04x %04x%04x%04x%04x  %*s", 
		 i, 
		 npx.reg[i].sign, npx.reg[i].exponent,
                 npx.reg[i].sig3, npx.reg[i].sig2,
                 npx.reg[i].sig1, npx.reg[i].sig0,
		 -(cols - 51), dstr);
	fs_put (1, y++, buf);
      }
    sprintf (buf, "%*s", cols - 19, "");
    while (y < 17)
      fs_put (1, y++, buf);
    load_npx ();
  }

  {
    /* Highlight focus */
    screen_attr = screen_attr_focus << 8;

    switch (pane)
      {
      case 0: /* Code */
        highlight (screen, 1, pane_pos + 1, cols - 19);
        break;
      case 1: /* Registers */
        highlight (screen, cols - 17, pane_pos + 1, 12);
        break;
      case 2: /* Flags */
        highlight (screen, cols - 4, pane_pos + 1, 3);
        break;
      case 3: /* Breakpoints */
        highlight (screen, cols - 32, 18 + 2 * pane_pos, 31);
        highlight (screen, cols - 32, 19 + 2 * pane_pos, 31);
        break;
      case 4: /* Data */
        highlight (screen, 11 + 3 * (pane_pos & 7), 18 + (pane_pos >> 3), 2);
        highlight (screen, 38 + (pane_pos & 7), 18 + (pane_pos >> 3), 1);
        break;
      case 5: /* NPX */
        highlight (screen, 1, pane_pos + 1, cols - 19);
        break;
      }
  }

  put_screen (screen);
  free (screen);
}
/* ---------------------------------------------------------------------- */
static void
initialize ()
{
  int i;

  debug_screen_p = 1;
  cols = ScreenCols ();
  rows = ScreenRows ();

  user_screen_save = get_screen ();

  pane = 0;
  pane_pos = 0;
  for (i = 0; i < PANECOUNT; i++)
    pane_positions[i] = 0;
  data_dump_origin = areas[A_data].first_addr;
  code_dump_origin = a_tss.tss_eip;
  code_pane_active = 1;
  npx_pane_active = 0;

  switch (ScreenMode ())
    {
    case 2:
    case 7:
      screen_attr_normal   = (A_black << 4) + A_grey;
      screen_attr_focus    = (A_grey  << 4) + A_black;
      screen_attr_break    = (A_black << 4) + A_white;
      screen_attr_message  = (A_grey  << 4) + A_white;
      break;
    default:
      screen_attr_normal   = (A_cyan  << 4) + A_blue;
      screen_attr_focus    = (A_blue  << 4) + A_white;
      screen_attr_break    = (A_red   << 4) + A_white;
      screen_attr_message  = (A_green << 4) + A_white;
    }
  redraw (1);
  message (0, "Sally Full Screen Debugger");
}
/* ---------------------------------------------------------------------- */
static void
code_pane_command (int key)
{
  int b;

  switch (key)
    {
    case K_EUp:
      if (pane_pos > 0)
        pane_pos--;
      else
        code_pane_goto (code_skip (code_dump_origin, -1));
      break;
    case K_EDown:
      if (pane_pos < 15)
        pane_pos++;
      else
        code_dump_origin =
          code_pane_pos[0] == code_pane_pos[1]
            ? code_pane_pos[2] : code_pane_pos[1];
      break;
    case K_EPageUp:
      code_dump_origin = code_skip (code_dump_origin, -16);
      break;
    case K_EPageDown:
      code_dump_origin = code_dump_last + 1;
      break;
    case K_Control_Left:
    case K_Control_ELeft:
      code_dump_origin--;
      break;
    case K_Control_Right:
    case K_Control_ERight:
      code_dump_origin++;
      break;
    case K_F2:
      b = get_breakpoint (code_pane_pos[pane_pos]);
      if (b != -1)
        reset_breakpoint (b);
      else
        set_breakpoint (-1, code_pane_pos[pane_pos]);
      break;
    case K_F4:
      set_breakpoint (0, code_pane_pos[pane_pos]);
      step (2);
      reset_breakpoint (0);
      break;
    case K_Control_O:
      code_pane_goto (a_tss.tss_eip);
      break;
    case K_Control_N:
      a_tss.tss_eip = code_pane_pos[pane_pos];
      break;
    case K_Control_G:
      {
        int ok, res;
        res = read_eval (&ok, "");
        if (ok)
          code_pane_goto (res);
        break;
      }
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
register_pane_command (int key)
{
  switch (key)
    {
    case K_ELeft:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_ERight:
    case K_EDown:
      if (pane_pos < 15) pane_pos++;
      break;
    default:
      if (key >= ' ' && key < 127)
        {
          int res, ok;
          char s[2];

          s[0] = key; s[1] = '\0';
          res = read_eval (&ok, key == '=' ? "" : s);
          if (ok)
            switch (pane_pos)
              {
              case  0: a_tss.tss_eax = res; break;
              case  1: a_tss.tss_ebx = res; break;
              case  2: a_tss.tss_ecx = res; break;
              case  3: a_tss.tss_edx = res; break;
              case  4: a_tss.tss_esi = res; break;
              case  5: a_tss.tss_edi = res; break;
              case  6: a_tss.tss_ebp = res; break;
              case  7: a_tss.tss_esp = res; break;
              case  8: a_tss.tss_cs  = res & 0xffff; break;
              case  9: a_tss.tss_ds  = res & 0xffff; break;
              case 10: a_tss.tss_es  = res & 0xffff; break;
              case 11: a_tss.tss_fs  = res & 0xffff; break;
              case 12: a_tss.tss_gs  = res & 0xffff; break;
              case 13: a_tss.tss_ss  = res & 0xffff; break;
              case 14: a_tss.tss_eip = res; break;
              case 15: a_tss.tss_eflags
                = (a_tss.tss_eflags & ~0xed5) | (res & 0xed5); break;
              }
        }
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
flag_pane_command (int key)
{
  static unsigned flagbits[] =
    { 0x0001, 0x0004, 0x0010, 0x0040, 0x0080, 0x0200, 0x0400, 0x0800 };

  switch (key)
    {
    case K_Space:
    case '+':
    case '-':
    case 't':
    case 'T':
      a_tss.tss_eflags ^= flagbits[pane_pos];
      break;
    case '1':
    case 's':
    case 'S':
      a_tss.tss_eflags |= flagbits[pane_pos];
      break;
    case '0':
    case 'r':
    case 'R':
      a_tss.tss_eflags &= ~flagbits[pane_pos];
      break;
    case K_ELeft:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_ERight:
    case K_EDown:
      if (pane_pos < 7) pane_pos++;
      break;
    default: /* Development... */
      message (1, "You hit key number 0x%x", key);
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
breakpoint_pane_command (int key)
{
  int b = pane_pos + 1;

  switch (key)
    {
    case K_ELeft:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_ERight:
    case K_EDown:
      if (pane_pos < 2) pane_pos++;
      break;
    case K_EDelete:
    case K_Control_C:
      reset_breakpoint (b);
      break;
    case K_Control_G:
    case K_Return:
      if ((edi.dr[7] & (3 << (b * 2))) != 0)
        code_pane_goto (edi.dr[b] - edi.app_base);
      break;
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
data_pane_command (int key)
{
  word32 v = data_dump_origin + pane_pos;

  switch (key)
    {
    case K_EUp:
      if (pane_pos >= 8)
        pane_pos -= 8;
      else
        data_dump_origin -= 8;
      break;
    case K_EDown:
      if (v + 8 <= data_dump_last)
        pane_pos += 8;
      else
        data_dump_origin += 8;
      break;
    case K_ELeft:
      if (pane_pos > 0)
        pane_pos--;
      else
        data_dump_origin -= 8, pane_pos = 7;
      break;
    case K_ERight:
      if (v < data_dump_last)
        pane_pos++;
      else
        data_dump_origin += 8, pane_pos -= 7;
      break;
    case K_EPageUp:
      data_dump_origin -= (data_dump_last + 1 - data_dump_origin);
      break;
    case K_EPageDown:
      data_dump_origin = data_dump_last + 1;
      break;
    case K_Control_G:
      {
        int ok, res;
        res = read_eval (&ok, "");
        if (ok)
          {
            data_dump_origin = res & -8;
            pane_pos = res & 7;
          }
        break;
      }
    default:
      if (key >= ' ' && key < 127 && valid_addr (v, 1))
        {
          int res, ok;
          char s[2];

          s[0] = key; s[1] = '\0';
          res = read_eval (&ok, key == '=' ? "" : s);
          if (ok)
            write_child (v, &res, 1);
        }
    }
  redraw (0);
}
/* ---------------------------------------------------------------------- */
static void
npx_pane_command (int key)
{
  int reg = pane_pos - 3, rotreg;
  int regp = (reg >= 0);

  save_npx ();
  rotreg = (reg + (npx.status >> 11)) & 7;
  switch (key)
    {
    case K_ELeft:
    case K_EUp:
      if (pane_pos > 0) pane_pos--;
      break;
    case K_ERight:
    case K_EDown:
      if (pane_pos < 10) pane_pos++;
      break;
    case K_Control_C:
    case K_Control_E:
    case K_EDelete:
      if (regp)
	{
	  npx.tag |= 3 << (rotreg * 2);
	  memset (&npx.reg[reg], 0, sizeof (NPXREG));
	}
      break;
    case K_Control_Z:
      if (regp)
	{
	  npx.tag = (npx.tag | (3 << (rotreg * 2))) - (2 << (rotreg * 2));
	  memset (&npx.reg[reg], 0, sizeof (NPXREG));
	}
      break;
    case K_Control_N:
      if (regp)
	npx.reg[reg].sign = !npx.reg[reg].sign;
      break;
    default:
      if (regp && key >= ' ' && key < 127)
        {
          int exp;
          char s[2], *endp, *p;
	  double d;

          s[0] = key; s[1] = '\0';
          if (!read_string (key == '=' ? "" : s) && read_buffer[0] != '\0')
	    {
	      p = read_buffer;
	      while (*p == ' ')
		p++;
	      d = strtod (p, &endp);
	      if (*p != '\0' && *endp)
		message (1, "Expression not understood");
	      else
		{
		  npx.tag &= ~(3 << (rotreg * 2));
		  if (d == 0.0)
		    {
		      npx.tag |= (1 << (rotreg * 2));
		      memset (&npx.reg[reg], 0, sizeof (NPXREG));
		    }
		  else
		    {
		      d = frexp (d, &exp);
		      npx.reg[reg].sign = (d >= 0) ? 0 : (d = -d, 1);
		      npx.reg[reg].exponent = exp + 16382;
		      d = ldexp (d, 16);
		      d -= (double) (npx.reg[reg].sig3 = floor (d));
		      d = ldexp (d, 16);
		      d -= (double) (npx.reg[reg].sig2 = floor (d));
		      d = ldexp (d, 16);
		      d -= (double) (npx.reg[reg].sig1 = floor (d));
		      d = ldexp (d, 16);
		      npx.reg[reg].sig0 = floor (d);
		    }
		}
	    }
        }
    }
  load_npx ();
  redraw (0);
}
/* ---------------------------------------------------------------------- */
void
debugger(void)
{
  int oldpane = -1;

  edi.dr[0] = edi.dr[1] = edi.dr[2] = edi.dr[3] = edi.app_base;
  main_entry = syms_name2val ("_main");
  first_step = !undefined_symbol;
  initialize ();
  can_longjmp = 1;
  setjmp (debugger_jmpbuf);

  while (1)
    {
      int key;

      if (pane == oldpane)
	pane_positions[pane] = pane_pos;
      else
	{
	  pane_pos = pane_positions[pane];
	  oldpane = pane;
	  redraw (0);
	}
      key = getxkey ();
      switch (key)
        {
        case K_Tab:
	  switch (pane)
	    {
	    case 1:
	    case 2:
	    case 3:
	      pane++;
	      break;
	    case 4:
	      if (code_pane_active)
		pane = 0;
	      else if (npx_pane_active)
		pane = 5;
	      break;
	    default:
	      pane = 1;
	      break;
	    }
          break;
        case K_BackTab:
	  switch (pane)
	    {
	    case 2:
	    case 3:
	    case 4:
	      pane--;
	      break;
	    case 1:
	      if (code_pane_active)
		pane = 0;
	      else if (npx_pane_active)
		pane = 5;
	      break;
	    default:
	      pane = 4;
	      break;
	    }
          break;
	case K_Alt_N:
	  code_pane_active = 0;
	  npx_pane_active = 1;
	  if (pane < 1 || pane > 4)
	    pane = 5;
	  oldpane = -1;
	  break;
	case K_Alt_C:
	  code_pane_active = 1;
	  npx_pane_active = 0;
	  if (pane < 1 || pane > 4)
	    pane = 0;
	  oldpane = -1;
	  break;
        case K_Alt_X:
          user_screen ();
          return;
        case K_Alt_F5:
          user_screen ();
          (void) getxkey ();
          debug_screen ();
          break;
        case K_F7:
          step (0);
          break;
        case K_F8:
          step (1);
          break;
        case K_F9:
          step (2);
          break;
        default:
          switch (pane)
            {
            case 0: code_pane_command (key); break;
            case 1: register_pane_command (key); break;
            case 2: flag_pane_command (key); break;
            case 3: breakpoint_pane_command (key); break;
            case 4: data_pane_command (key); break;
            case 5: npx_pane_command (key); break;
            }
        }
    }
}
/* ---------------------------------------------------------------------- */
