/*
 * scantest.c: use poly_scan() for Gouraud shading demo.
 *    Splotches N Gouraud triangles on the screen.
 *                           
 * Paul Heckbert	Dec 1989
 *                              
 * (Modified July 1993 by David Max to test true color routines)
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "poly.h"
#include "hixcolor.h"
#include "graphics.h"

#define XMAX 640   	/* image width */
#define YMAX 480   	/* image height */

#define FRANDOM() ((rand()&32767)/32767.)   /* random number between 0 and 1 */

void pixelproc();

void main(int argc, char **argv)
{
    int i,j;
    Poly p;
    int N; 
    Window win = {0, 0, XMAX-1, YMAX-1}; 	/* screen clipping window */

    if (argc<2 || (N = atoi(argv[1])) < 1) {
   	fprintf(stderr, "Usage: scantest N \n");
   	exit(1);
    }

    GrSetMode_24();  /* Start 24 bit graphics mode */

    for (j=0; j<N; j++) {
       p.n = 3;
       for (i=0; i<p.n; i++) {
           p.vert[i].sx = FRANDOM()*XMAX;
           p.vert[i].sy = FRANDOM()*YMAX;
           p.vert[i].sz = 0.0;
           p.vert[i].r = FRANDOM();        /* random vertex colors, for kicks */
           p.vert[i].g = FRANDOM();
           p.vert[i].b = FRANDOM();
       }
       /* interpolate sx, sy, sz, r, g, and b in poly_scan */
       p.mask = POLY_MASK(sx) | POLY_MASK(sy) | POLY_MASK(sz) |
           POLY_MASK(r) | POLY_MASK(g) | POLY_MASK(b);
   
       poly_scan(&p, &win, pixelproc);     /* scan convert! */

       if (kbhit()) break;
    }

    while(!kbhit()) /* Wait for Keypress */;
    getch();
    GrSetMode(GR_default_text);
}   

void pixelproc(int x, int y, Poly_vert *point)	/* called at each pixel by poly_scan */
{
    GrPlot24(x, y, (pixel_24)(255*point->r), 
                   (pixel_24)(255*point->g), 
                   (pixel_24)(255*point->b));
}
