;--------------------------------------------------------------------------
; This is file VGAEXT.INC
;
; Copyright (C) 1991 DJ Delorie, 24 Kirsten Ave, Rochester NH 03867-2954
; Copyright (C) 1992 Csaba Biegl, 820 Stirrup Dr, Nashville, TN 37221
; Copyright (C) 1993 Hartmut Schirmer, Feldstr. 118, 2300 Kiel 1, Germany
;
; This file is distributed under the terms listed in the document
; "copying.dj", available from DJ Delorie at the address above.
; A copy of "copying.dj" should accompany this file; if not, a copy
; should be available from where this file was obtained.  This file
; may not be distributed without a verbatim copy of "copying.dj".
;
; This file is distributed WITHOUT ANY WARRANTY; without even the implied
; warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
;--------------------------------------------------------------------------
;
;  These are the special VGA modes taken from STDVGA.ASM
;
;  Supplied modes are
;
;    80x28x16     textmode
;    80x50x16     textmode
;    720x350x16   nice to emulate a Hercules Graphics card
;    720x480x16   enhanced resolution on every VGA
;    320x240x256  enhanced 256 color mode (Mode X)
;    320x240x256  enhanced 256 color mode (Mode X)
;    320x240x256  enhanced 256 color mode (Mode X)
;
;  Usage as follows:
;
;  1. define all modes you want to include, eg.:
;
;     req_T28         equ    1        ; enable  80x28x16 text mode
;     req_T50         equ    0        ; disable 80x50x16 text mode
;     req_3516        equ    1        ; enable  720x350x16 graphics mode
;     req_4816        equ    1        ; enable  720x480x16 graphics mode
;     req_lowX        equ    1        ; enable  320x240/320x400/360x480x256
;     req_hiX         equ    0        ; disable high resolution mode X support
;
;  2. set up the mode tables (again from et4000x.asm) :
;     text_mode_table         label word
;         ...
;         if req_T28
;             dw      80,     28,     16,     val_T28
;         endif
;         if req_T50
;             dw      80,     50,     16,     val_T50
;         endif
;         ...
;     graphics_mode_table     label word
;         ...
;         if req_3516
;             dw      720,    350,    16,     val_3516
;         endif
;         if req_4816
;             dw      720,    480,    16,     val_4816
;         endif
;         ...
;             dw      320,    200,    256,    013h +  val_hiX ; Mode X like
;         if req_lowX
;             dw      320,    240,    256,    val_lowXa       ; Mode X
;             dw      320,    400,    256,    val_lowXb       ; Mode X
;             dw      360,    480,    256,    val_lowXc       ; Mode X
;         endif
;         ...
;
;  3. next the 'TABLE OF SPECIAL SETUP PROCEDURES' are defined and
;     vgaext.inc is includes directly after all needed driver specific
;     table entries:
;
;     special_setup_table label word    ; use exactly this label!
;             dw      offset init1      ; special init routines
;             dw      offset init2      ; for modes not within
;             dw      offset init3      ; vgaext.inc *FIRST*
;     include vgaext.inc                ; load VGA non BIOS modes
;
;  4. Now define the driver specific initialisation routines
;     init1 proc near
;     init1 endp
;     init2 proc near
;     init2 endp
;     init3 proc near
;     init3 endp
;
;  5. Complete the driver ...
;
; ----------------------------------------------------------------------------

    if req_T28
	dw      offset  VGA_28row_mode_set
val_T28 equ     ($ - offset special_setup_table)*128
    endif
    if req_T50
	dw      offset  VGA_50row_mode_set
val_T50 equ     ($ - offset special_setup_table)*128
    endif
    if req_3516
	dw      offset  VGA_720x350x16
val_3516 equ    010h +  ($ - offset special_setup_table)*128
    endif
    if req_4816
	dw      offset  VGA_720x480x16
val_4816 equ    012h +  ($ - offset special_setup_table)*128
    endif
    if req_lowX
	dw      offset  VGA_320x240x256
val_lowXa equ   013h +  ($ - offset special_setup_table)*128
	dw      offset  VGA_320x400x256
val_lowXb equ   013h +  ($ - offset special_setup_table)*128
	dw      offset  VGA_360x480x256
val_lowXc equ   013h +  ($ - offset special_setup_table)*128
    endif
    if req_hiX
	dw      offset  VGA_256_to_X
val_hiX equ     ($ - offset special_setup_table)*128
    endif

    if req_T50
;
; Routine to set up VGA 50 row mode
;
VGA_50row_mode_set      proc    near
	mov     ax,[di+6]               ; BIOS mode number
	xor     ah,ah                   ; set up basic BIOS text mode (x25)
	int     10h
	xor     bx,bx
	mov     ax,1112h                ; load 8x8 font
	int     10h
	ret
VGA_50row_mode_set      endp
    endif


    if req_T28
;
; Routine to set up VGA 28 row mode
;
VGA_28row_mode_set      proc    near
	mov     ax,[di+6]               ; BIOS mode number
	xor     ah,ah                   ; set up basic BIOS text mode (x25)
	xor     bx,bx
	mov     ax,1111h                ; load 8x14 font
	int     10h
	ret
VGA_28row_mode_set      endp
    endif


    if req_lowX
;
; Routine to set up VGA 320 x 240 x 256 color mode  (Mode X)
;
VGA_320x240x256  proc    near
	mov     si, offset reg_320x240
	jmp     mode_X

reg_320x240:
	dw      03c4h
	db      000h
	dw      03c5h
	db      001h
	dw      03c2h
	db      0e3h
	dw      03c5h
	db      003h
	dw      03c4h
	db      004h
	dw      03c5h
	db      006h
	dw      0000h

	db      05Fh,04Fh,050h,082h,054h,080h,00Dh,03Eh
	db      000h,041h,000h,000h,000h,000h,000h,000h
	db      0EAh,0ACh,0DFh,028h,000h,0E7h,006h,0E3h

VGA_320x240x256 endp

;
; Routine to set up VGA 320 x 400 x 256 color mode  (Mode X)
;
VGA_320x400x256  proc    near
	mov     si, offset reg_320x400
	jmp     mode_X

reg_320x400:
	dw      03c4h
	db      004h
	dw      03c5h
	db      006h
	dw      0000h

	db      05Fh,04Fh,050h,082h,054h,080h,0BFh,01Fh
	db      000h,040h,000h,000h,000h,000h,000h,000h
	db      09Ch,08Eh,08Fh,028h,000h,096h,0B9h,0E3h
VGA_320x400x256 endp

;
; Routine to set up VGA 360 x 480 x 256 color mode  (Mode X)
;
VGA_360x480x256  proc    near
	mov     si, offset reg_360x480
mode_X:
	push    si
	mov     ax, 12h                 ; clear video memory
	int     10h
	pop     si
	mov     bx, GRD_8_X_PLANES
	jmp     non_BIOS_mode

reg_360x480:
	dw      03c4h
	db      000h
	dw      03c5h
	db      001h
	dw      03c2h
	db      0e7h
	dw      03c5h
	db      003h
	dw      03c4h
	db      004h
	dw      03c5h
	db      006h
	dw      0000h

	db      06Bh,059h,05Ah,08Eh,05Eh,08Ah,00Dh,03Eh
	db      000h,040h,000h,000h,000h,000h,000h,000h
	db      0EAh,0ACh,0DFh,02Dh,000h,0E7h,006h,0E3h

VGA_360x480x256 endp

    endif ;req_lowX

    if req_3516
;
; Routine to set up VGA 720 x 480 x 16 color mode
;
VGA_720x480x16  proc near
	mov     si, offset reg_720x480
	mov     bx, GRD_4_PLANES
	jmp     non_BIOS_mode

reg_720x480:
	dw      03c4h
	db      000h
	dw      03c5h
	db      001h
	dw      03c2h
	db      0e7h
	dw      03c5h
	db      003h
	dw      0000h

	db      06bh, 059h, 05ah, 08eh, 05eh, 08ah, 00dh, 03eh
	db      000h, 040h, 000h, 000h, 000h, 000h, 000h, 000h
	db      0eah, 0ach, 0dfh, 02dh, 00fh, 0e7h, 006h, 0e3h

VGA_720x480x16  endp
    endif

    if req_4816
;
; Routine to set up VGA 720 x 350 x 16 color mode
;
VGA_720x350x16  proc near
	mov     si, offset reg_720x350
	mov     bx, GRD_4_PLANES
	jmp     non_BIOS_mode

reg_720x350:
	dw      03c4h
	db      000h
	dw      03c5h
	db      001h
	dw      03c2h
	db      0a7h
	dw      03c5h
	db      003h
	dw      0000h

	db      06bh, 059h, 05ah, 08eh, 05eh, 08ah, 0bfh, 01fh
	db      000h, 040h, 000h, 000h, 000h, 000h, 000h, 000h
	db      083h, 085h, 05dh, 02dh, 00fh, 063h, 0bah, 0e3h

VGA_720x350x16  endp
    endif


    if req_3516 or req_4816 or req_lowX
;
; Set up a non BIOS mode
;
; entry : si -> register values table
;         bx -> GRX memory mode flags
;
non_BIOS_mode proc near
	push    bx
	push    si
	mov     ax,[di+6]                       ; BIOS mode number
	xor     ah,ah                           ; set up basic BIOS
	int     10h                             ; mode
	pop     si

nBm_a:  mov     ax, word ptr cs:[si]            ; First part of the table
	inc     si                              ; is port.w, value.b
	inc     si
	or      ax, ax                          ; port.w == 0 terminates
	je      nBm_b                           ; at exit: si -> CRTC values
	mov     dx, ax
	mov     al, byte ptr cs:[si]
	inc     si
	out     dx, al
	jmp     short nBm_a

nBm_b:  mov     dx, 03d4h                       ; Second part are the 0x18
	mov     al, 011h                        ; CRTC values
	out     dx, al                          ; Set CRTC writable
	inc     dx
	in      al, dx
	and     al, 07fh
	out     dx, al

	dec     dx                              ; and reprogram the controller
	mov     cl, 0
nBm_c:  mov     al, cl
	inc     cl
	cmp     al, 18h
	je      short nBm_d
	out     dx, al
	inc     dx
	mov     al, byte ptr cs:[si]
	inc     si
	out     dx, al
	dec     dx
	jmp     short nBm_c

nBm_d:  pop     bx
	or      bx, mode_W
	ret

non_BIOS_mode endp
    endif


    if req_hiX
;
; Routine to set up a high resolution VGA 256 color X mode
; Mode setting is done calling the BIOS, memory organization
; is switched afterwards
;
VGA_256_to_X proc    near
	mov     ax,[di+6]                       ; BIOS mode number
	xor     ah,ah                           ; set up linear 256 color
	int     10h                             ; mode
	mov     dx, 03C4h                       ; switch to mode X
	mov     al, 4
	out     dx, al
	inc     dx
	in      al, dx
	and     al, 0F7h
	or      al, 004h
	out     dx, al
	mov     bx, GRD_8_X_PLANES              ; return mode X like
	ret
VGA_256_to_X endp
    endif
