#    Copyright (C) 1989, 1990 Aladdin Enterprises.  All rights reserved.
#    Distributed by Free Software Foundation, Inc.
#
# This file is part of Ghostscript.
#
# Ghostscript is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
# to anyone for the consequences of using it or for whether it serves any
# particular purpose or works at all, unless he says so in writing.  Refer
# to the Ghostscript General Public License for full details.
#
# Everyone is granted permission to copy, modify and redistribute
# Ghostscript, but only under the conditions described in the Ghostscript
# General Public License.  A copy of this license is supposed to have been
# given to you along with Ghostscript so you can know your rights and
# responsibilities.  It should be in a file named COPYING.  Among other
# things, the copyright notice and this notice must be preserved on all
# copies.

# makefile for Ghostscript, Unix/gcc/X11 configuration.

# ------------------------------- Options ------------------------------- #

####### The following are the only parts of the file you should need to edit.

# ------ Generic options ------ #

# Define the default directory/ies for the runtime
# initialization and font files.  Separate multiple directories with a :.
# `pwd` means use the directory in which the 'make' is being done.

GS_LIB_DEFAULT=`pwd`:`pwd`/fonts

# Choose generic configuration options.

# -DDEBUG
#	includes debugging features (-Z switch) in the code.
#	  Code runs substantially slower even if no debugging switches
#	  are set.
# -DNOPRIVATE
#	makes private (static) procedures and variables public,
#	  so they are visible to the debugger and profiler.
#	  No execution time or space penalty.

GENOPT=

# ------ Platform-specific options ------ #

# Define the name of the C compiler.

CC=gcc

# Define the other compilation flags.
# Add -DBSD4_2 for 4.2bsd systems.
# Add -DUSG (GNU convention) or -DSYSV for System V.
# We don't include -ansi, because this gets in the way of the platform-
#   specific stuff that <math.h> typically needs.

CFLAGS=-g -O

# Define platform flags for ld.
# Most Unix systems accept -X, but some don't.
# Sun OS4.n needs -Bstatic.

LDPLAT=-X

# Define any extra libraries to link into the executable.
# The default is for X Windows.

EXTRALIBS= -lX11

# --------------------------- Choice of devices --------------------------- #

# Choose the device(s) to include.  See gdevs.mak for details.

DEVICES=x11
DEVICE_OBJS=$(x11_)

# ---------------------------- End of options --------------------------- #

# Define the name of the makefile -- used in dependencies.

MAKEFILE=unix-gcc.mak

# Define the extensions for the object and executable files.

OBJ=o
XE=

# Define the ANSI-to-K&R dependency.  (gcc accepts ANSI syntax.)

AK=

# Define the directory separator and shell quote string.

DS=/
Q=\"

# Define the compilation rules.

CCFLAGS=$(GENOPT) $(CFLAGS)

.c.o:
	$(CC) $(CCFLAGS) -c $*.c

CCA=$(CC) $(CCFLAGS) -c
CCNA=$(CCA)
CCINT=$(CCA)

# --------------------------- Generic makefile ---------------------------- #

# The remainder of the makefile (ghost.mak, gdevs.mak, and unixtail.mak)
# is generic.  tar_gs concatenates all these together.
