#    Copyright (C) 1989, 1990, 1991 Aladdin Enterprises.  All rights reserved.
#    Distributed by Free Software Foundation, Inc.
#
# This file is part of Ghostscript.
#
# Ghostscript is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
# to anyone for the consequences of using it or for whether it serves any
# particular purpose or works at all, unless he says so in writing.  Refer
# to the Ghostscript General Public License for full details.
#
# Everyone is granted permission to copy, modify and redistribute
# Ghostscript, but only under the conditions described in the Ghostscript
# General Public License.  A copy of this license is supposed to have been
# given to you along with Ghostscript so you can know your rights and
# responsibilities.  It should be in a file named COPYING.  Among other
# things, the copyright notice and this notice must be preserved on all
# copies.

# makefile for Ghostscript device drivers.

# -------------------------------- Catalog ------------------------------- #

# It is possible to build Ghostscript with an arbitrary collection of
# device drivers, although many drivers are supported only on a subset
# of the target platforms.  The currently available drivers are:

# Displays:
#	bgi	Borland Graphics Interface   [MS-DOS only]
#	ega	EGA display   [MS-DOS only]
#	ega_bios  EGA display using BIOS calls (very slow)   [MS-DOS only]
#	mdb10	EIZO MDB-10 display (1024 x 768)   [MS-DOS only]
#	sonyfb	Sony Microsystems monochrome display   [Sony only]
#	sunview  SunView window system   [SunOS only]
#	vga	VGA display   [MS-DOS only]
#	x11	X Windows version 11, release >=3   [Unix only]
# (NOTE: no more than one MS-DOS display driver may be included in
#   any build.)
# Printers:
#	bj10e	Canon BubbleJet BJ10e  [MS-DOS only]
#	deskjet  H-P DeskJet   [MS-DOS & Unix]
#	epson	Epson dot matrix printers   [MS-DOS only]
#	laserjet  H-P LaserJet   [MS-DOS & Unix]
#	nwp533  Sony Microsystems NWP533 laser printer   [Sony only]
# ###	paintjet  H-P PaintJet color printer [Unix, tested on SunOS only]
# ###	  (NOT INCLUDED IN THIS RELEASE)

# If you add drivers, it would be nice if you kept each list
# in alphabetical order.

# Each platform-specific makefile should contain a pair of lines of the form
#	DEVICES=<dev1> ... <devn>
#	DEVICE_OBJS=$(<dev1>_) ... $(<devn>_)
# where dev1 ... devn are the devices to be included in the build.
# dev1 will be used as the default device.  Don't forget the _s!
# On MS-DOS platforms, the first of these lines must appear *before*,
# and the second *after*, the lines
#	(!)include gdevs.mak
#	(!)include ghost.mak
# in the makefile.  On Unix systems, the two device definition lines
# may appear anywhere in the makefile.
#
# The executable must be linked with all the files named in DEVICE_OBJS.
# On MS-DOS platforms, this is done by constructing a file called gdevs.tl,
# to get around the limit on the length of a DOS command line.

# ---------------------------- End of catalog ---------------------------- #

# If you want to add a new device driver, the examples below should be
# enough of a guide to the correct form for the makefile rules.

# All device drivers depend on the following
# (note that we include some indirect dependencies explicitly):

GDEV=$(AK) gs.h gx.h gsmatrix.h gxbitmap.h gxdevice.h

###### ------------------- MS-DOS display devices ------------------- ######

GDEVPCFB=gdevpcfb.h $(GDEV)

gdevegaa.$(OBJ): gdevegaa.asm

### -------------------------- The EGA device -------------------------- ###

ETEST=ega.$(OBJ) trace.$(OBJ) $(ega_)
ega.exe: $(ETEST) libc$(MM).tl
	tlink /m /l $(LIBDIR)\c0$(MM) @ega.tl @libc$(MM).tl

ega.$(OBJ): ega.c $(GDEV)

ega_=gdevega.$(OBJ) gdevegaa.$(OBJ)
ega.dev: $(ega_)
	.$(DS)gssetdev ega.dev $(ega_)

gdevega.$(OBJ): gdevpcfb.c gdevega.h $(GDEVPCFB)
	cp gdevega.h gdevxxfb.h
	$(CCNA) gdevpcfb.c
	rm gdevxxfb.h
	cp gdevpcfb.obj gdevega.obj

ega_bios_=gdevegab.$(OBJ) gdevegaa.$(OBJ)
ega_bios.dev: $(ega_bios_)
	.$(DS)gssetdev ega_bios.dev $(ega_bios_)

gdevegab.$(OBJ): gdevegab.c $(GDEVPCFB)
	$(CCNA) $(D_EGA) gdevegab.c

### -------------------------- The VGA device -------------------------- ###

vga_=gdevvga.$(OBJ) gdevegaa.$(OBJ)
vga.dev: $(vga_)
	.$(DS)gssetdev vga.dev $(vga_)

gdevvga.$(OBJ): gdevpcfb.c gdevvga.h $(GDEVPCFB)
	cp gdevvga.h gdevxxfb.h
	$(CCNA) gdevpcfb.c
	rm gdevxxfb.h
	cp gdevpcfb.obj gdevvga.obj

### ---------------------- The EIZO MDB-10 device ---------------------- ###

mdb10_=gdevmd10.$(OBJ) gdevegaa.$(OBJ)
mdb10.dev: $(mdb10_)
	.$(DS)gssetdev mdb10.dev $(mdb10_)

gdevmd10.$(OBJ): gdevpcfb.c gdevmd10.h $(GDEVPCFB)
	cp gdevmd10.h gdevxxfb.h
	$(CCNA) gdevpcfb.c
	rm gdevxxfb.h
	cp gdevpcfb.obj gdevmd10.obj

###### --------- The BGI (Borland Graphics Interface) device -------- ######

bgi_=gdevbgi.$(OBJ) egavga.$(OBJ)
bgi.dev: $(bgi_)
	.$(DS)gssetdev bgi.dev $(bgi_)

gdevbgi.$(OBJ): gdevbgi.c $(GDEV)
	$(CCNA) gdevbgi.c

egavga.$(OBJ): $(COMPDIR)\egavga.bgi
	bgiobj egavga

###### --------------- Memory-buffered printer devices --------------- ######

gvirtmem.$(OBJ): gvirtmem.c gvirtmem.h

# Virtual memory test program -- requires VMDEBUG

gvm.exe: gvirtmem.$(OBJ) trace.$(OBJ)
	tlink /m /l $(LIBDIR)\c0$(MM) gvirtmem trace,gvm,gvm,$(LIBDIR)\c$(MM)

PDEVH=$(GDEV) gxdevmem.h gvirtmem.h gdevprn.h

# We use CCNA only because the MS-DOS version of the driver
# refers to stdprn, which is non-ANSI.
gdevprn.$(OBJ): gdevprn.c $(PDEVH)
	$(CCNA) gdevprn.c

### ----------------- The Canon BubbleJet BJ10e device ----------------- ###

bj10e_=gdevbj10.$(OBJ) gvirtmem.$(OBJ) gdevprn.$(OBJ)
bj10e.dev: $(bj10e_)
	.$(DS)gssetdev bj10e.dev $(bj10e_)

gdevbj10.$(OBJ): gdevbj10.c $(PDEVH)

### ------------------ The H-P DeskJet printer device ------------------ ###

# Note that this shares code with the LaserJet device (below).

deskjet_=gdevdjet.$(OBJ) gvirtmem.$(OBJ) gdevprn.$(OBJ)
deskjet.dev: $(deskjet_)
	.$(DS)gssetdev deskjet.dev $(deskjet_)

gdevdjet.$(OBJ): gdevdjet.c $(PDEVH)
	$(CCA) -DLASER=0 gdevdjet.c

### ----------------- The generic Epson printer device ----------------- ###

epson_=gdevepsn.$(OBJ) gvirtmem.$(OBJ) gdevprn.$(OBJ)
epson.dev: $(epson_)
	.$(DS)gssetdev epson.dev $(epson_)

gdevepsn.$(OBJ): gdevepsn.c $(PDEVH)

### ------------------ The H-P LaserJet printer device ----------------- ###

# Note that this shares code with the DeskJet device (above).

laserjet_=gdevljet.$(OBJ) gvirtmem.$(OBJ) gdevprn.$(OBJ)
laserjet.dev: $(laserjet_)
	.$(DS)gssetdev laserjet.dev $(laserjet_)

gdevljet.$(OBJ): gdevdjet.c $(PDEVH)
	cp gdevdjet.c gdevljet.c
	$(CCA) -DLASER=1 gdevljet.c
	rm gdevljet.c

### ------------ The H-P PaintJet color printer device ----------------- ###
### Note: this driver was contributed by users:                          ###
###       please contact marc@vlsi.polymtl.ca if you have questions.     ###

### This driver was not ready to be included in release 2.0.
### Do not attempt to use it: the files are not included in the fileset.

PJETH=$(GDEV) gdevprn.h gdevpjet.h

paintjet_=gdevpjet.$(OBJ)
paintjet.dev: $(paintjet_)
	.$(DS)gssetdev paintjet.dev $(paintjet_)

gdevpjet.$(OBJ): gdevpjet.c $(PJETH)

###### ------------------ Sony frame buffer device ----------------- ######

sonyfb_=gdevsnfb.$(OBJ) gvirtmem.$(OBJ) gdevprn.$(OBJ)
sonyfb.dev: $(sonyfb_)
	.$(DS)gssetdev sonyfb.dev $(sonyfb_)

gdevsnfb.$(OBJ): gdevsnfb.c $(PDEVH)
	$(CCA) gdevsnfb.c

###### ----------------- Sony NWP533 printer device ----------------- ######

nwp533_=gdevn533.$(OBJ) gvirtmem.$(OBJ) gdevprn.$(OBJ)
nwp533.dev: $(nwp533_)
	.$(DS)gssetdev nwp533.dev $(nwp533_)

gdevn533.$(OBJ): gdevn533.c $(PDEVH)
	$(CCA) gdevn533.c

###### --------------------- The SunView device --------------------- ######

sunview_=gdevsun.$(OBJ)
sunview.dev: $(sunview_)
	.$(DS)gssetdev sunview.dev $(sunview_)

gdevsun.$(OBJ): gdevsun.c $(GDEV)

###### ----------------------- The X11 device ----------------------- ######

x11_=gdevx.$(OBJ)
x11.dev: $(x11_)
	.$(DS)gssetdev x11.dev $(x11_)

### Note: if the X11 client header libraries are not on /usr/include/X11,
### you may have to change the compilation line below to add a -I switch.
### For example, if the header files are in /usr/local/X/include/X11,
### you must add the switch -I/usr/local/X/include.
gdevx.$(OBJ): gdevx.c $(GDEV) gdevx.h
	$(CCA) -I/usr/local/X/include gdevx.c
