#    Copyright (C) 1989, 1990 Aladdin Enterprises.  All rights reserved.
#    Distributed by Free Software Foundation, Inc.
#
# This file is part of Ghostscript.
#
# Ghostscript is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY.  No author or distributor accepts responsibility
# to anyone for the consequences of using it or for whether it serves any
# particular purpose or works at all, unless he says so in writing.  Refer
# to the Ghostscript General Public License for full details.
#
# Everyone is granted permission to copy, modify and redistribute
# Ghostscript, but only under the conditions described in the Ghostscript
# General Public License.  A copy of this license is supposed to have been
# given to you along with Ghostscript so you can know your rights and
# responsibilities.  It should be in a file named COPYING.  Among other
# things, the copyright notice and this notice must be preserved on all
# copies.

# Generic makefile for Ghostscript.
# The platform-specific makefiles 'include' this file.
# They define the following symbols:
#	GS_LIB_DEFAULT - the default directory/ies for searching for the
#		initialization and font files at run time.
#	Q - the string that causes the shell to pass a " to a program
#		(" on MS-DOS, \" on Unix).
#	XE - the extension for executable files (e.g., null or .exe).
#	OBJ - the extension for relocatable object files (e.g., o or obj).
#	CCA - the ANSI C invocation for normal compilation.
#	CCNA - the non-ANSI C invocation for files that on some platforms
#		include in-line assembly code or other non-standard
#		constructs.  Currently this is needed on every file that
#		includes sstorei.h, and a few files that use Turbo C
#		"short pointer" constructs.
#	CCINT - the C invocation for compiling the main interpreter module,
#		normally the same as CCNA: this is needed because the
#		Borland compiler generates *worse* code for this module
#		(but only this module) when optimization (-O) is turned on.
#	AK - if source files must be converted from ANSI to K&R syntax,
#		this is ansi2knr$(XE); if not, it is null.
#	DS - the directory separator (/ or \).
# The platform-specific makefiles must also include rules for creating
#   ansi2knr$(XE) and genarch$(XE) from the corresponding .c files,
#   and for making arch.h by executing genarch$(XE).  (This
#   shouldn't really be necessary, but Turbo C and Unix C treat the -o
#   switch slightly differently (Turbo C requires no following space,
#   Unix C requires a following space), and I haven't found a way to capture
#   the difference in a macro; also, Unix requires ./ because . may not be
#   in the search path, whereas MS-DOS always looks in the current
#   directory first.)

default: gs$(XE)

test: gt$(XE)

clean:
	rm -f *.$(OBJ) *.a core gmon.out
	rm -f *.dev gdevs.h gdevs.tl
	rm -f t _temp_* _temp_*.* libc*.tl *.map *.sym
	rm -f ansi2knr$(XE) genarch$(XE) arch.h gs$(XE)

# Note: Unix uses malloc.h and memory.h;
# Turbo C uses alloc.h, stdlib.h, and mem.h.
# 4.2bsd uses strings.h; other systems use string.h.
# gcc on VMS doesn't have a math.h.
# We handle this by using local include files called
# malloc_.h, math_.h, memory_.h, and string_.h
# that perform appropriate indirection.

# Auxiliary programs

arch.h: genarch$(XE)
	.$(DS)genarch

# -------------------------------- Library -------------------------------- #

GX=$(AK) std.h gx.h
GXERR=$(GX) gserrors.h

###### High-level facilities

gschar.$(OBJ): gschar.c $(GXERR) \
  gxfixed.h gxarith.h gxmatrix.h gzdevice.h gxdevmem.h gxfont.h gxchar.h gstype1.h gzpath.h gzcolor.h gzstate.h

gscolor.$(OBJ): gscolor.c $(GXERR) \
  gxfixed.h gxmatrix.h gxdevice.h gzstate.h gzcolor.h gzht.h

gscoord.$(OBJ): gscoord.c $(GXERR) \
  gxfixed.h gxmatrix.h gzdevice.h gzstate.h gscoord.h

gsdevice.$(OBJ): gsdevice.c $(GXERR) \
  gxfixed.h gxmatrix.h gxbitmap.h gxdevmem.h gzstate.h gzdevice.h

gsfile.$(OBJ): gsfile.c $(GXERR) \
  gsmatrix.h gxdevice.h gxdevmem.h

gsfont.$(OBJ): gsfont.c $(GXERR) \
  gxdevice.h gxfixed.h gxmatrix.h gxfont.h gxfdir.h gzstate.h

gsimage.$(OBJ): gsimage.c $(GXERR) \
  arch.h gxfixed.h gxarith.h gxmatrix.h gspaint.h gzcolor.h gzdevice.h gzpath.h gzstate.h gximage.h

gsim2out.$(OBJ): gsim2out.c $(GXERR) \
  gsstate.h gsmatrix.h gscoord.h gxfixed.h gxtype1.h

gsline.$(OBJ): gsline.c $(GXERR) \
  gxfixed.h gxmatrix.h gzstate.h gzline.h

gsmatrix.$(OBJ): gsmatrix.c $(GXERR) \
  gxfixed.h gxarith.h gxmatrix.h

gsmisc.$(OBJ): gsmisc.c $(GX)

gspaint.$(OBJ): gspaint.c $(GX) \
  gxfixed.h gxmatrix.h gspaint.h gzpath.h gzstate.h gzdevice.h gximage.h

gspath.$(OBJ): gspath.c $(GXERR) \
  gxfixed.h gxmatrix.h gxpath.h gzstate.h

gspath2.$(OBJ): gspath2.c $(GXERR) \
  gspath.h gxfixed.h gxmatrix.h gzstate.h gzpath.h gzdevice.h

gsstate.$(OBJ): gsstate.c $(GXERR) \
  gxfixed.h gxmatrix.h gzstate.h gzcolor.h gzdevice.h gzht.h gzline.h gzpath.h

gstype1.$(OBJ): gstype1.c $(GXERR) \
  gxarith.h gxfixed.h gxmatrix.h gxchar.h gxdevmem.h gxtype1.h gxfont1.h gzstate.h gzdevice.h gzpath.h

###### Low-level facilities

gxcache.$(OBJ): gxcache.c $(GX) \
  gserrors.h gxfixed.h gxmatrix.h gspaint.h gzdevice.h gzcolor.h gxdevmem.h gxfont.h gxfdir.h gxchar.h gzstate.h gzpath.h

gxcolor.$(OBJ): gxcolor.c $(GX) \
  gxfixed.h gxmatrix.h gxdevice.h gzcolor.h gzht.h

gxdither.$(OBJ): gxdither.c $(GX) \
  gxfixed.h gxmatrix.h gzstate.h gzdevice.h gzcolor.h gzht.h

gxdraw.$(OBJ): gxdraw.c $(GX) \
  gxfixed.h gxmatrix.h gxbitmap.h gzcolor.h gzdevice.h gzstate.h

gxfill.$(OBJ): gxfill.c $(GXERR) \
  gxfixed.h gxmatrix.h gxdevice.h gzcolor.h gzpath.h gzstate.h

gxht.$(OBJ): gxht.c $(GXERR) \
  gxfixed.h gxmatrix.h gxbitmap.h gzstate.h gzcolor.h gzdevice.h gzht.h

gxpath.$(OBJ): gxpath.c $(GXERR) \
  gxfixed.h gzpath.h

gxpath2.$(OBJ): gxpath2.c $(GXERR) \
  gxfixed.h gxarith.h gzpath.h

gxstroke.$(OBJ): gxstroke.c $(GXERR) \
  gxfixed.h gxarith.h gxmatrix.h gzstate.h gzcolor.h gzdevice.h gzline.h gzpath.h

###### The "memory" device

gdevmem.$(OBJ): gdevmem.c $(AK) \
  gs.h arch.h gxbitmap.h gsmatrix.h gxdevice.h gxdevmem.h
	$(CCNA) gdevmem.c

###### Files dependent on the set of installed devices.
###### Generating gdevs.h also generates gdevs.tl.

gdevs.h: gdevs.mak $(MAKEFILE) makefile
	.$(DS)gsconfig $(DEVICES)

gdevs.$(OBJ): gdevs.c $(AK) gdevs.h

###### On Unix, we pre-link all of the library except the back end.
###### On MS-DOS, we have to do the whole thing at once.

LIB=gschar.$(OBJ) gscolor.$(OBJ) gscoord.$(OBJ) gsdevice.$(OBJ) \
 gsfile.$(OBJ) gsfont.$(OBJ) gsimage.$(OBJ) gsim2out.$(OBJ) \
 gsline.$(OBJ) gsmatrix.$(OBJ) gsmisc.$(OBJ) \
 gspaint.$(OBJ) gspath.$(OBJ) gspath2.$(OBJ) gsstate.$(OBJ) gstype1.$(OBJ) \
 gxcache.$(OBJ) gxcolor.$(OBJ) gxdither.$(OBJ) gxdraw.$(OBJ) gxfill.$(OBJ) \
 gxht.$(OBJ) gxpath.$(OBJ) gxpath2.$(OBJ) gxstroke.$(OBJ) \
 gdevmem.$(OBJ) gdevs.$(OBJ)

# ------------------------------ Interpreter ------------------------------ #

###### Utilities

GH=$(AK) ghost.h

ialloc.$(OBJ): ialloc.c $(AK) std.h alloc.h
	$(CCNA) ialloc.c

idebug.$(OBJ): idebug.c $(GH) name.h

idict.$(OBJ): idict.c $(GH) alloc.h errors.h name.h store.h dict.h

iinit.$(OBJ): iinit.c $(GH) dict.h oper.h store.h

iname.$(OBJ): iname.c $(GH) alloc.h errors.h name.h store.h

iscan.$(OBJ): iscan.c $(GH) arch.h alloc.h dict.h errors.h name.h store.h stream.h scanchar.h

iutil.$(OBJ): iutil.c $(GH) errors.h alloc.h oper.h store.h gsmatrix.h gxdevice.h gzcolor.h

stream.$(OBJ): stream.c $(AK) std.h stream.h scanchar.h gxfixed.h gstype1.h

### Level 2 utilities

i2num.$(OBJ): i2num.c $(GH) arch.h errors.h stream.h i2num.h i2btoken.h

i2scan.$(OBJ): i2scan.c $(GH) arch.h errors.h alloc.h store.h stream.h i2bseq.h i2btoken.h i2num.h

###### Non-graphics operators

OP=$(GH) errors.h oper.h

zarith.$(OBJ): zarith.c $(OP) store.h

zarray.$(OBJ): zarray.c $(OP) alloc.h store.h sstorei.h
	$(CCNA) zarray.c

zcontrol.$(OBJ): zcontrol.c $(OP) estack.h store.h sstorei.h
	$(CCNA) zcontrol.c

zdict.$(OBJ): zdict.c $(OP) dict.h store.h

zfile.$(OBJ): zfile.c $(OP) alloc.h stream.h store.h gsmatrix.h gxdevice.h gxdevmem.h

zgeneric.$(OBJ): zgeneric.c $(OP) dict.h estack.h store.h

zmath.$(OBJ): zmath.c $(OP) store.h

zmisc.$(OBJ): zmisc.c $(OP) alloc.h dict.h store.h gstype1.h gxfixed.h

zpacked.$(OBJ): zpacked.c $(OP) store.h

zrelbit.$(OBJ): zrelbit.c $(OP) store.h sstorei.h dict.h
	$(CCNA) zrelbit.c

zstack.$(OBJ): zstack.c $(OP) store.h sstorei.h
	$(CCNA) zstack.c

zstring.$(OBJ): zstring.c $(OP) alloc.h store.h stream.h

ztype.$(OBJ): ztype.c $(OP) dict.h name.h stream.h store.h

zvmem.$(OBJ): zvmem.c $(OP) alloc.h state.h store.h gsmatrix.h gsstate.h

### Level 2 non-graphics operators

z2bseq.$(OBJ): z2bseq.c $(OP) stream.h name.h i2btoken.h i2bseq.h

###### Graphics operators

zchar.$(OBJ): zchar.c $(OP) gxmatrix.h gschar.h gstype1.h gxdevice.h gxfixed.h gxfont.h gxfont1.h gzpath.h gzstate.h alloc.h dict.h font.h estack.h state.h store.h

zcolor.$(OBJ): zcolor.c $(OP) alloc.h gsmatrix.h gsstate.h state.h store.h

zdevice.$(OBJ): zdevice.c $(OP) alloc.h state.h gsmatrix.h gsstate.h gxdevice.h store.h

zfont.$(OBJ): zfont.c $(OP) gsmatrix.h gxdevice.h gxfont.h gxfont1.h alloc.h font.h dict.h name.h state.h store.h

zgstate.$(OBJ): zgstate.c $(OP) alloc.h gsmatrix.h gsstate.h state.h store.h

zht.$(OBJ): zht.c $(OP) alloc.h estack.h gsmatrix.h gsstate.h state.h store.h

zmatrix.$(OBJ): zmatrix.c $(OP) gsmatrix.h state.h gscoord.h store.h

zpaint.$(OBJ): zpaint.c $(OP) alloc.h estack.h gsmatrix.h gspaint.h state.h store.h

zpath.$(OBJ): zpath.c $(OP) gsmatrix.h gspath.h state.h store.h

zpath2.$(OBJ): zpath2.c $(OP) alloc.h estack.h gspath.h state.h store.h

### Level 2 graphics operators

z2path.$(OBJ): z2path.c $(OP) state.h stream.h i2num.h

###### Linking

INT=ialloc.$(OBJ) idebug.$(OBJ) idict.$(OBJ) iinit.$(OBJ) iname.$(OBJ) \
 interp.$(OBJ) iscan.$(OBJ) iutil.$(OBJ) stream.$(OBJ) \
 zarith.$(OBJ) zarray.$(OBJ) zcontrol.$(OBJ) zdict.$(OBJ) zfile.$(OBJ) \
 zgeneric.$(OBJ) zmath.$(OBJ) zmisc.$(OBJ) zpacked.$(OBJ) zrelbit.$(OBJ) \
 zstack.$(OBJ) zstring.$(OBJ) ztype.$(OBJ) zvmem.$(OBJ) \
 zchar.$(OBJ) zcolor.$(OBJ) zfont.$(OBJ) zdevice.$(OBJ) zgstate.$(OBJ) \
 zht.$(OBJ) zmatrix.$(OBJ) zpaint.$(OBJ) zpath.$(OBJ) zpath2.$(OBJ)

INT2=i2num.$(OBJ) \
 z2bseq.$(OBJ) \
 z2path.$(OBJ)

# ----------------------------- Main program ------------------------------ #

# Utilities shared between platforms

gsmain.$(OBJ): gsmain.c $(GX) \
  gsmatrix.h gxdevice.h

# Library test program driver

gt.$(OBJ): gt.c $(GX) \
  gsmatrix.h gsstate.h gscoord.h gspaint.h gspath.h gxdevice.h

# Interpreter main program

interp.$(OBJ): interp.c $(GH) \
  errors.h name.h dict.h oper.h store.h sstorei.h stream.h
	$(CCINT) interp.c

gs.$(OBJ): gs.c $(GH) alloc.h store.h stream.h $(MAKEFILE)
	$(CCA) -DGS_LIB_DEFAULT=$(Q)$(GS_LIB_DEFAULT)$(Q) gs.c
