/* rmdir -- remove directories
   Copyright (C) 1990 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * MS-DOS port (c) 1990 by Thorsten Ohl, td12@ddagsi3.bitnet
 *
 * To this port, the same copying conditions apply as to the
 * original release.
 *
 * IMPORTANT:
 * This file is not identical to the original GNU release!
 * You should have received this code as patch to the official
 * GNU release.
 *
 * MORE IMPORTANT:
 * This port comes with ABSOLUTELY NO WARRANTY.
 *
 * $Header: e:/gnu/fileutil/RCS/rmdir.c'v 1.3.0.2 90/06/29 00:47:09 tho Stable $
 */

/* Usage: rmdir [-p] [+path] dir...

   Options:
   -p, +path		Remove any parent dirs that are explicitly mentioned
			in an argument, if they become empty after the
			argument file is removed.

   David MacKenzie <djm@ai.mit.edu>  */

#include <stdio.h>
#include <sys/types.h>
#include "system.h"
#include "getopt.h"

#ifdef MSDOS
#include <direct.h>
extern void main (int argc, char * *argv);
extern void error (int status, int errnum, char *message, ...);
static void remove_parents (char *path);
static void strip_trailing_slashes (char *path);
static void usage (void);
#endif /* MSDOS */

#ifdef STDC_HEADERS
#include <errno.h>
#include <stdlib.h>
#else
extern int errno;
#endif

void remove_parents ();
void error ();
void strip_trailing_slashes ();
void usage ();

/* If nonzero, remove empty parent directories. */
int empty_paths;

/* The name this program was run with. */
char *program_name;

struct option longopts[] =
{
  {"path", 0, &empty_paths, 1},
  {NULL, 0, NULL, 0}
};

void
main (argc, argv)
     int argc;
     char **argv;
{
  int errors = 0;
  int optc;
  int ind;

  program_name = argv[0];
#ifdef MSDOS
  strlwr (program_name);
#endif /* MSDOS */
  empty_paths = 0;

  while ((optc = getopt_long (argc, argv, "p", longopts, &ind)) != EOF)
    {
      switch (optc)
	{
	case 0:			/* Long option. */
	  break;
	case 'p':
	  empty_paths = 1;
	  break;
	default:
	  usage ();
	}
    }

  if (optind == argc)
    usage ();
  
  for (; optind < argc; ++optind)
    {
#ifdef MSDOS
      strlwr (argv[optind]);
#endif /* MSDOS */

      strip_trailing_slashes (argv[optind]);
      if (rmdir (argv[optind]) != 0)
	{
	  error (0, errno, "%s", argv[optind]);
	  errors = 1;
	}
      else if (empty_paths)
	remove_parents (argv[optind]);
    }

  exit (errors);
}

/* Remove any empty parent directories of `path'.
   Replaces '/' characters in `path' with NULs. */

void
remove_parents (path)
     char *path;
{
  char *slash;

  do
    {
      slash = rindex (path, '/');
      if (slash == NULL)
	break;
      /* Remove any characters after the slash, skipping any extra
	 slashes in a row. */
      while (slash > path && *slash == '/')
	--slash;
      slash[1] = 0;
    }
  while (rmdir (path) == 0);
}

/* Remove trailing slashes from PATH; they cause some system calls to fail. */

void
strip_trailing_slashes (path)
     char *path;
{
  int last;

  last = strlen (path) - 1;
  while (last > 0 && path[last] == '/')
    path[last--] = '\0';
}

void
usage ()
{
  fprintf (stderr, "Usage: %s [-p] [+path] dir...\n",
	   program_name);
  exit (1);
}
