/* $Id$ */
/*
 * go32info.c
 *
 * Display the memory usage of GO32 into an X window,
 */
   
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include <X11/bitmaps/stipple>

#include <stdio.h>
#include <sys/types.h>
#include <go32.h>
#include <dpmi.h>
#include <time.h>

extern time_t rawclock(void);

static  Display *mydisplay;
static  Window   mywindow;
static  GC mygc;
static  XGCValues mygcvalues;
static  XEvent myevent;
static  XSizeHints myhint;
static  int myscreen;
static  unsigned long myforeground, mybackground;
static  Cursor mycursor;
static  Pixmap mypixmap;

static	char go32_version[30] = "GO32 info";
static  char *go32_runmode;
static	time_t next_time = 0L;

static int
main (int argc, char *argv[])
{
  mydisplay = XOpenDisplay ("");
  myscreen = DefaultScreen (mydisplay);
  mybackground = WhitePixel (mydisplay, myscreen);
  myforeground = BlackPixel (mydisplay, myscreen);
  
  myhint.x = DisplayWidth(mydisplay, myscreen) - 250;
  myhint.y = 10;
  myhint.width = 225;
  myhint.height = 75;
  myhint.min_width = myhint.width;
  myhint.min_height = myhint.height;
  myhint.flags = PPosition | PSize | PMinSize;

  mywindow = XCreateSimpleWindow (mydisplay,
				  DefaultRootWindow (mydisplay),
				  myhint.x, myhint.y,
				  myhint.width, myhint.height,
				  5, myforeground, mybackground);
  XSetStandardProperties (mydisplay, mywindow, go32_version, go32_version,
			  None, argv, argc, &myhint);

  mycursor = XCreateFontCursor (mydisplay, XC_left_ptr);
  XDefineCursor (mydisplay, mywindow, mycursor);

  mygc = XCreateGC (mydisplay, mywindow, 0, 0);
  mypixmap = XCreateBitmapFromData (mydisplay, mywindow, 
				    stipple_bits, stipple_width, stipple_height);

  mygcvalues.foreground = myforeground;
  mygcvalues.background = mybackground;
  mygcvalues.fill_style = FillStippled;
  mygcvalues.stipple = mypixmap;
  XChangeGC (mydisplay, mygc,
	     GCForeground | GCBackground | GCFillStyle | GCStipple, &mygcvalues);

  XSelectInput (mydisplay, mywindow,
		ButtonPressMask
		| KeyPressMask
		| ExposureMask
		| EnterWindowMask
		| LeaveWindowMask);

  XMapRaised (mydisplay, mywindow);
  return 0;
}

static void
drawinfo (void)
{
  char buffer[100];
  int pm, vm, y;
  time_t t;

  pm = _go32_dpmi_remaining_physical_memory();
  vm = _go32_dpmi_remaining_virtual_memory();
  y = 0;
  
  t = time (0L);
  sprintf (buffer, "%s", ctime (&t));
  XDrawImageString (mydisplay, mywindow, mygc, 5, (y += 20),
		    buffer, strlen (buffer));
  sprintf (buffer, "Physical memory = %5d kb ", pm / 1024);
  XDrawImageString (mydisplay, mywindow, mygc, 5, (y += 20),
		    buffer, strlen (buffer));
  sprintf (buffer, "Virtual memory  = %5d kb ", vm / 1024);
  XDrawImageString (mydisplay, mywindow, mygc, 5, (y += 20),
		    buffer, strlen (buffer));
}

static void
events (void)
{
  while (XPending (mydisplay)) {

    XNextEvent (mydisplay, &myevent);

    switch (myevent.type) 
      {
      case Expose:
	if (myevent.xexpose.count == 0)
	  drawinfo ();
	break;
      }
  }
}

static void
term (void)
{  
  XFreeGC (mydisplay, mygc);
  XDestroyWindow (mydisplay, mywindow);
  XCloseDisplay (mydisplay);
  /* exit (0); */
}

void
go32info(void)
{
  time_t curr_time;
  int argc;
  char *argv[2];
  char dpmi_version[20];

  if (next_time == 0L) {

    switch (_go32_info_block.run_mode) {
    case _GO32_RUN_MODE_RAW:
      go32_runmode = "raw";
      break;
    case _GO32_RUN_MODE_XMS:
      go32_runmode = "xms";
      break;
    case _GO32_RUN_MODE_VCPI:
      go32_runmode = "vcpi";
      break;
    case _GO32_RUN_MODE_DPMI:
      sprintf (dpmi_version, "dpmi %d.%d%d",
	       (_go32_info_block.run_mode_info>>8)&0xff,
	       (_go32_info_block.run_mode_info>>4)&0x0f,
	       (_go32_info_block.run_mode_info>>0)&0x0f);
      go32_runmode = dpmi_version;
      break;
    default:
      go32_runmode = "undef";
      break;
    }

    sprintf (go32_version, "GO32 %s info", go32_runmode);

    argc = 0;
    argv[argc++] = "go32info";
    main (argc, argv);
  }

  /* Handle events */
  events ();

  /* Update about every second */
  curr_time = rawclock();
  if (curr_time > next_time)
    {
      /* Number of bios timer ticks for one second = 18.2046 ticks */
      next_time = curr_time + 18;
      drawinfo ();
    }
}
