/* $Id: property.c 1.2 1994/01/15 02:09:52 ulrich Exp $ */
/*
 * property.c
 *
 * X library functions for window properties.
 */
#include "Xlibemu.h"

typedef struct _WPropertyRec {
  unsigned char *value;
  Atom property;
  Atom type;
  int format;
  int nitems;
} _WProperty;

typedef struct {
  Window window;
  int size;
  int next;
  _WProperty *props;
} _WWindowProperty;

typedef struct {
  int size;
  int next;
  _WWindowProperty *windows;
} _WWindowPropertyList;

_WWindowPropertyList _PropertyList = { 0 };


int XGetWindowProperty(
    Display*		display,
    Window		w,
    Atom		property,
    long		long_offset,
    long		long_length,
    Bool		delete,
    Atom		req_type,
    Atom*		actual_type_return,
    int*		actual_format_return,
    unsigned long*	nitems_return,
    unsigned long*	bytes_after_return,
    unsigned char**	prop_return)
{
  int i;
  _WWindowProperty *wp;
  _WProperty *pp;

  for (i = _PropertyList.next; --i >= 0; ) {
    if (_PropertyList.windows[i].window == w)
      break;
  }
  if (i < 0) {
  not_found:
    *actual_type_return = None;
    *actual_format_return = 0;
    *bytes_after_return = 0;
    *nitems_return = 0;
    return -1;
  }
  wp = &_PropertyList.windows[i];
  for (i = wp->next; --i >= 0; ) {
    if (wp->props[i].property == property)
      break;
  }
  if (i < 0) {
    goto not_found;
  }
  pp = &wp->props[i];
  if (pp->type == req_type || req_type == AnyPropertyType) {
    int n, i, t, l, a;
    unsigned char *data;
    
    n = (pp->format/8) * pp->nitems;
    i = 4 * long_offset;
    t = n - i;
    l = min (t, 4 * long_length);
    if (l < 0) {
      _WError (display, (XID) w, BadValue, X_GetProperty);
      goto not_found;
    }
    a = n - (i + l);

    data = (char *) Xmalloc (l+1);
    memcpy (data, (char *) pp->value + i, l);
    data[l] = '\0';

    *actual_type_return = pp->type;
    *actual_format_return = pp->format;
    *nitems_return = l / (pp->format/8);
    *bytes_after_return = a;
    *prop_return = data;
  }
  else {
    *actual_type_return = pp->type;
    *actual_format_return = pp->format;
    *bytes_after_return = (pp->format/8) * pp->nitems;
    *nitems_return = 0;
  }
  if (delete == True && *bytes_after_return == 0) {
    wp->next--;
    if (pp < &wp->props[wp->next]) {
      Xfree (pp->value);
      memcpy (pp, pp+1, (char *) &wp->props[wp->next] - (char *)pp);
    }
    {
      XEvent xe;

      xe.xproperty.type = PropertyNotify;
      xe.xproperty.send_event = False;
      xe.xproperty.window = w;
      xe.xproperty.atom = property;
      xe.xproperty.time = _WCurrentTime ();
      xe.xproperty.state = PropertyDelete;
      _WDispatchEvent (&xe);
    }
  }
  return Success;
}

Atom *XListProperties(
    Display*		display,
    Window		w,
    int*		num_prop_return)
{
  int i;
  _WWindowProperty *wp;
  Atom *ap;

  for (i = _PropertyList.next; --i >= 0; ) {
    if (_PropertyList.windows[i].window == w)
      break;
  }
  if (i < 0) {
    *num_prop_return = 0;
    return NULL;
  }
  wp = &_PropertyList.windows[i];
  ap = (Atom *) Xmalloc (wp->next * sizeof(Atom));
  for (i = wp->next; --i >= 0; ) {
    ap[i] = wp->props[i].property;
  }
  *num_prop_return = wp->next;
  return ap;
}

int XChangeProperty(
    Display*		display,
    Window		w,
    Atom		property,
    Atom		type,
    int			format,
    int			mode,
    _Xconst unsigned char*	data,
    int			nelements)
{
  int i, nbytes, nbytes_p;
  unsigned char *value;
  _WWindowProperty *wp;
  _WProperty *pp;

  for (i = _PropertyList.next; --i >= 0; ) {
    if (_PropertyList.windows[i].window == w)
      break;
  }
  if (i < 0) {
    /* add window w to window list */
    if (_PropertyList.next >= _PropertyList.size) {
      _PropertyList.size += _PropertyList.size/2 + 1;
      _PropertyList.windows = (_WWindowProperty *)
	Xrealloc (_PropertyList.windows,
		  _PropertyList.size * sizeof(_WWindowProperty));
    }
    wp = &_PropertyList.windows[_PropertyList.next++];
    wp->window = w;
    wp->size = 0;
    wp->next = 0;
    wp->props = 0;
  }
  else {
    wp = &_PropertyList.windows[i];
  }
  for (i = wp->next; --i >= 0; ) {
    if (wp->props[i].property == property)
      break;
  }
  if (i < 0) {
    /* create the property */
    if (wp->next >= wp->size) {
      wp->size += wp->size/2 + 1;
      wp->props = (_WProperty *) 
	Xrealloc (wp->props,
		  wp->size * sizeof (_WProperty));
    }
    pp = &wp->props[wp->next++];
    nbytes = nelements * (format/8);
    pp->value = (char *) Xmalloc (nbytes);
    memcpy (pp->value, data, nbytes);
    pp->property = property;
    pp->type = type;
    pp->format = format;
    pp->nitems = nelements;
  }
  else {
    pp = &wp->props[i];

    if (mode != PropModeReplace
	&& (pp->type != type || pp->format != format)) {
      _WError (display, (XID) w, BadMatch, X_ChangeProperty);
      return 0;
    }

    nbytes = nelements * (format/8);
    nbytes_p = pp->nitems * (format/8);

    switch (mode) {
    case PropModeReplace:
      Xfree (pp->value);
      pp->format = format;
      pp->type = type;
      pp->value = (unsigned char *) Xmalloc (nbytes);
      pp->nitems = nelements;
      memcpy (pp->value, data, nbytes);
      break;
    case PropModePrepend:
      value = (unsigned char *) Xmalloc (nbytes + nbytes_p);
      memcpy (value, data, nbytes);
      memcpy (value + nbytes, pp->value, nbytes_p);
      Xfree (pp->value);
      pp->value = value;
      pp->nitems += nelements;
      break;
    case PropModeAppend:
      value = (unsigned char *) Xmalloc (nbytes_p + nbytes);
      memcpy (value, pp->value, nbytes_p);
      memcpy (value + nbytes_p, data, nbytes);
      Xfree (pp->value);
      pp->value = value;
      pp->nitems += nelements;
      break;
    }
  }
  if (w != None) {
    XEvent xe;
    
    xe.xproperty.type = PropertyNotify;
    xe.xproperty.send_event = False;
    xe.xproperty.window = w;
    xe.xproperty.atom = property;
    xe.xproperty.time = _WCurrentTime ();
    xe.xproperty.state = PropertyNewValue;
    _WDispatchEvent (&xe);
  }
  return 1;
}

int XRotateWindowProperties(
    Display*		display,
    Window		w,
    Atom*		properties,
    int			num_prop,
    int			npositions)
{
  return 0;
}

int XDeleteProperty(
    Display*		display,
    Window		w,
    Atom		property)
{
  int i;
  _WWindowProperty *wp;
  _WProperty *pp;

  for (i = _PropertyList.next; --i >= 0; ) {
    if (_PropertyList.windows[i].window == w)
      break;
  }
  if (i < 0) {
    return 0;
  }
  wp = &_PropertyList.windows[i];
  for (i = wp->next; --i >= 0; ) {
    if (wp->props[i].property == property)
      break;
  }
  if (i < 0) {
    return 0;
  }
  pp = &wp->props[i];
  wp->next--;
  if (pp < &wp->props[wp->next]) {
    Xfree (pp->value);
    memcpy (pp, pp+1, (char *) &wp->props[wp->next] - (char *)pp);
  }
  {
    XEvent xe;
    
    xe.xproperty.type = PropertyNotify;
    xe.xproperty.send_event = False;
    xe.xproperty.window = w;
    xe.xproperty.atom = property;
    xe.xproperty.time = _WCurrentTime ();
    xe.xproperty.state = PropertyDelete;
    _WDispatchEvent (&xe);
  }
  return 0;
}
