/* $Id: winint.c 1.1 1994/02/20 17:59:15 ulrich Exp ulrich $ */
/*
 * Xlibemu internal window managment functions.
 */
#include "Xlibemu.h"

Bool
_WCheckWindow (Display *dpy, Window window, int request_code)
{
  if (window == None
      || (window->upper_sibling &&
	  window->upper_sibling->parent != window->parent)
      || (window->lower_sibling &&
	  window->lower_sibling->parent != window->parent)) {
    _WError (dpy, (XID) window, BadWindow, request_code);
    return 0;
  }
  return 1;
}

/* Window tree functions */

void
_WTreeDelete (Window w)
{
  if (w->upper_sibling)
    w->upper_sibling->lower_sibling = w->lower_sibling;
  else if (w->parent)
    w->parent->top_child = w->lower_sibling;

  if (w->lower_sibling)
    w->lower_sibling->upper_sibling = w->upper_sibling;
  else if (w->parent)
    w->parent->bottom_child = w->upper_sibling;

  w->parent = None;
  w->upper_sibling = None;
  w->lower_sibling = None;
}

void
_WTreeInsertTop (Window parent, Window window)
{
  if (parent == None) return;
  window->parent = parent;
  window->upper_sibling = None;
  window->lower_sibling = parent->top_child;
  if (parent->top_child)
    parent->top_child->upper_sibling = window;
  else
    parent->bottom_child = window;
  parent->top_child = window;
}

void
_WTreeInsertBot (Window parent, Window window)
{
  if (parent == None) return;
  window->parent = parent;
  window->lower_sibling = None;
  window->upper_sibling = parent->bottom_child;
  if (parent->bottom_child)
    parent->bottom_child->lower_sibling = window;
  else
    parent->top_child = window;
  parent->bottom_child = window;
}

void
_WTreeInsertAbove (Window sibling, Window window)
{
  window->parent = sibling->parent;
  window->lower_sibling = sibling;
  window->upper_sibling = sibling->upper_sibling;
  if (sibling->upper_sibling)
    sibling->upper_sibling->lower_sibling = window;
  else
    window->parent->top_child = window;
  sibling->upper_sibling = window;
}

void
_WTreeInsertBelow (Window sibling, Window window)
{
  window->parent = sibling->parent;
  window->lower_sibling = sibling->lower_sibling;
  window->upper_sibling = sibling;
  if (sibling->lower_sibling)
    sibling->lower_sibling->upper_sibling = window;
  else
    window->parent->bottom_child = window;
  sibling->lower_sibling = window;
}

void
_WMapWindow (Display* dpy, Window window)
{
  int viewable;

  window->mapped = 1;
  viewable = (window->parent)
    ? window->parent->viewable
    : True;
  if (viewable != window->viewable) {
    window->viewable = viewable;
    /* set viewable for all subwindows depending on map state */
    _WSetViewable (window, viewable);
  }
  if (window->class == InputOutput) {
    if (window->parent) {
      _WSetVisibleSubwindowsInBox (window->parent, &window->border_port);
      _WInvalidateParent (window);
    }
    else {
      _WSetVisibleRegion (window);
      _WSetVisibleSubwindows (window);
    }
    _WDrawWindow (dpy, window, &window->border_port, True);
    _WDrawSubwindows (dpy, window, &window->border_port, True);
  }
}

void
_WUnmapWindow (Display* dpy, Window window)
{
  window->mapped = 0;
  window->viewable = 0;
  _WSetViewable (window, 0);	/* make all subwindows unviewable */
  if (window->parent != None &&
      window->class == InputOutput) {
    _WSetVisibleRegion (window->parent);
    _WSetVisibleSubwindowsInBox (window->parent, &window->border_port);
    _WDrawSiblingsBelow (dpy, window, True);
    _WDrawWindowBackground (dpy, window->parent, &window->border_port, True);
  }
}

void
_WDestroyWindow (Display *dpy, Window w)
{
  if (w->mapped) XUnmapWindow (dpy, w);

  _WDrawInvalid (w);	/* flush drawable cache */
  _WTreeDelete (w);

  w->event_mask = 0;
  w->do_not_propagate_mask = ~0;
  w->visibility = VisibilityFullyObscured;
  EMPTY_REGION(w->visible_region);

  if (w->context) {
    GrDestroyContext (w->context);
    w->context = NULL;
  }
#if 0
  /* The event queue may contain references to w */

  XDestroyRegion (w->visible_region);
  Xfree (w);
#endif
}
