/* $Id: debug.c 1.2 1994/01/15 02:08:36 ulrich Exp $ */
/*
 * debug.c
 *
 * Debugging functions for development only.
 */
#include "Xlibemu.h"
#include <stdio.h>

#ifdef DEBUG

static char *_EventNames[] = {
  "0",
  "1",
  "KeyPress",
  "KeyRelease",
  "ButtonPress",
  "ButtonRelease",
  "MotionNotify",
  "EnterNotify",
  "LeaveNotify",
  "FocusIn",
  "FocusOut",
  "KeymapNotify",
  "Expose",
  "GraphicsExpose",
  "NoExpose",
  "VisibilityNotify",
  "CreateNotify",
  "DestroyNotify",
  "UnmapNotify",
  "MapNotify",
  "MapRequest",
  "ReparentNotify",
  "ConfigureNotify",
  "ConfigureRequest",
  "GravityNotify",
  "ResizeRequest",
  "CirculateNotify",
  "CirculateRequest",
  "PropertyNotify",
  "SelectionClear",
  "SelectionRequest",
  "SelectionNotify",
  "ColormapNotify",
  "ClientMessage",
  "MappingNotify"
};

void
_XPrintEvent (Display *display, XEvent *event)
{
  char *name, buffer[10];

  if (event->xany.type >= LASTEvent)
    sprintf (buffer, "%d", event->xany.type), name = buffer;
  else
    name = _EventNames[event->xany.type];

  fprintf (stderr, "Event: %s
  Serial:\t%d
  Send Event:\t%d
  Window:\t%08x
  Event Mask:\t%08x
",
	   name, event->xany.serial, event->xany.send_event,
	   (int) event->xany.window,
	   event->xany.window->event_mask);

#if 0
  if (PointerGrab.active)
    fprintf (stderr, "  grab window=%08x mask=%08x\n",
	     (int) PointerGrab.grab_window,
	     PointerGrab.grab_window->event_mask);
#endif


}

void
_XPrintTree (Display *display, Window window, int level)
{
  int i;
  Window child;

  if (level == 0) {
    fprintf (stderr, "\nWindow Tree\n");
  }
  for (i = 0; i < level*2; i++) fputc (' ', stderr);

  fprintf (stderr, "%08x %dx%d%+d%+d p=%08x t=%08x b=%08x u=%08x d=%08x\n",
	  (int) window, window->width, window->height, window->x, window->y,
	  (int) window->parent,
	  (int) window->top_child,
	  (int) window->bottom_child,
	  (int) window->upper_sibling,
	  (int) window->lower_sibling);
  for (child = window->top_child;
       child != NULL;
       child = child->lower_sibling) {
    _XPrintTree (display, child, level + 1);
  }
  fflush (stderr);
}

void
_XPrintBox (Box *b)
{
  fprintf (stderr, "Box %08lx x1=%d y1=%d x2=%d y2=%d\n",
	   (int) b, b->x1, b->y1, b->x2, b->y2);
}

void
_XPrintRegion (Region rgn)
{
  int i;

  if (rgn->numRects == 0) {
    fprintf (stderr, "Region is empty\n");
    return;
  }
  fprintf (stderr, "Region Rectangles\n");
  for (i = 0; i < rgn->numRects; i++) {
    fprintf (stderr, "%d ", i);
    _XPrintBox (&rgn->rects[i]);
  }
  fprintf (stderr, "Region Extents\n");
  _XPrintBox (&rgn->extents);
  fflush (stderr);
}

#endif
