/* $Id: event.c 1.5 1994/01/27 01:38:37 ulrich Exp $ */
/*
 * event.c
 *
 * Internal event functions.
 */

#include "Xlibemu.h"
#ifdef DEBUG
#include <stdio.h>
#endif

extern _XQEvent *_qfree;

#define AllMotionMask \
(PointerMotionMask \
 | PointerMotionHintMask \
 | ButtonMotionMask \
 | Button1MotionMask \
 | Button2MotionMask \
 | Button3MotionMask \
 | Button4MotionMask \
 | Button5MotionMask)


int
_WEventButton (_WQueueEvent *ev)
{
  switch (ev->flags & M_BUTTON_CHANGE) {
  case M_LEFT_UP:
  case M_LEFT_DOWN:
    return Button1;
  case M_MIDDLE_UP:
  case M_MIDDLE_DOWN:
    return Button2;
  case M_RIGHT_UP:
  case M_RIGHT_DOWN:
    return Button3;
  }
  return 0;
}

unsigned int
_WButtonToMask (int button)
{
  switch (button) {
  case Button1: return Button1Mask;
  case Button2: return Button2Mask;
  case Button3: return Button3Mask;
  }
  return 0;
}

unsigned int
_WButtonAndModifierState (_WQueueEvent *event)
{
  return (((event->kbstat & KB_SHIFT) ? ShiftMask : 0)
	  | ((event->kbstat & KB_CAPSLOCK) ? LockMask : 0)
	  | ((event->kbstat & KB_CTRL) ? ControlMask : 0)
	  | ((event->kbstat & KB_ALT) ? Mod1Mask : 0)
	  | ((event->buttons & M_LEFT) ? Button1Mask : 0)
	  | ((event->buttons & M_MIDDLE) ? Button2Mask : 0)
	  | ((event->buttons & M_RIGHT) ? Button3Mask : 0));
}

/*
 * _XFlush - Flush the X request buffer.  If the buffer is empty, no
 * action is taken.  This routine correctly handles incremental writes.
 * This routine may have to be reworked if int < long.
 */
void
_XFlush (dpy)
	register Display *dpy;
{
}

static Bool _DoNotBlock = False;

int
_XEventsQueued (dpy, mode)
    register Display *dpy;
    int mode;
{
	if (mode == QueuedAfterFlush)
	{
	    _XFlush(dpy);
	    if (dpy->qlen)
		return(dpy->qlen);
	}
	if (_WQueuePendingEvent ()) {
	  _DoNotBlock = True;
	  _XReadEvents (dpy);
	  _DoNotBlock = False;
	}
	return(dpy->qlen);
}

/*
(let ((i 0))
 (while (< i 256)
  (progn
   (insert (format "0x%02x,\n" i))
   (setq i (+ i 1)))))
*/

static unsigned char scancode_remap[256] = {
0x00,
0x01,
0x02,
0x03,
0x04,
0x05,
0x06,
0x07,
0x08,
0x09,
0x0a,
0x0b,
0x0c,
0x0d,
0x0e,
0x0f,
0x10,
0x11,
0x12,
0x13,
0x14,
0x15,
0x16,
0x17,
0x18,
0x19,
0x1a,
0x1b,
0x1c,
0x1d,
0x1e,
0x1f,
0x20,
0x21,
0x22,
0x23,
0x24,
0x25,
0x26,
0x27,
0x28,
0x29,
0x2a,
0x2b,
0x2c,
0x2d,
0x2e,
0x2f,
0x30,
0x31,
0x32,
0x33,
0x34,
0x35,
0x36,
0x37,
0x38,
0x39,
0x3a,
0x3b,	/* F1 */
0x3c,
0x3d,
0x3e,
0x3f,
0x40,
0x41,
0x42,
0x43,
0x44,
0x45,
0x46,
0x47,
0x48,
0x49,
0x4a,
0x4b,
0x4c,
0x4d,
0x4e,
0x4f,
0x50,
0x51,
0x52,
0x53,
0x54 - (0x54 - 0x3b),	/* S-F1 */
0x55 - (0x54 - 0x3b),
0x56 - (0x54 - 0x3b),
0x57 - (0x54 - 0x3b),
0x58 - (0x54 - 0x3b),
0x59 - (0x54 - 0x3b),
0x5a - (0x54 - 0x3b),
0x5b - (0x54 - 0x3b),
0x5c - (0x54 - 0x3b),
0x5d - (0x54 - 0x3b),
0x5e - (0x5e - 0x3b),	/* C-F1 */
0x5f - (0x5e - 0x3b),
0x60 - (0x5e - 0x3b),
0x61 - (0x5e - 0x3b),
0x62 - (0x5e - 0x3b),
0x63 - (0x5e - 0x3b),
0x64 - (0x5e - 0x3b),
0x65 - (0x5e - 0x3b),
0x66 - (0x5e - 0x3b),
0x67 - (0x5e - 0x3b),
0x68 - (0x68 - 0x3b),	/* M-F1 */
0x69 - (0x68 - 0x3b),
0x6a - (0x68 - 0x3b),
0x6b - (0x68 - 0x3b),
0x6c - (0x68 - 0x3b),
0x6d - (0x68 - 0x3b),
0x6e - (0x68 - 0x3b),
0x6f - (0x68 - 0x3b),
0x70 - (0x68 - 0x3b),
0x71 - (0x68 - 0x3b),
0x72,
0x73 - (0x73 - 0x4b),	/* C-Left */
0x74 - (0x74 - 0x4d),	/* C-Right */
0x75 - (0x75 - 0x4f),	/* C-End */
0x76 - (0x76 - 0x51),	/* C-PageDown */
0x77 - (0x77 - 0x47),	/* C-Home */
0x78 - (0x78 - 0x02),	/* M-1 */
0x79 - (0x78 - 0x02),
0x7a - (0x78 - 0x02),
0x7b - (0x78 - 0x02),
0x7c - (0x78 - 0x02),
0x7d - (0x78 - 0x02),
0x7e - (0x78 - 0x02),
0x7f - (0x78 - 0x02),
0x80 - (0x78 - 0x02),
0x81 - (0x78 - 0x02),
0x82,
0x83,
0x84 - (0x84 - 0x49),	/* C-PageUp */
0x85,	/* F11 */
0x86,	/* F12 */
0x87 - (0x87 - 0x85),	/* S-F11 */
0x88 - (0x87 - 0x85),	/* S-F12 */
0x89 - (0x89 - 0x85),	/* C-F11 */
0x8a - (0x89 - 0x85),	/* C-F12 */
0x8b - (0x8b - 0x85),	/* M-F11 */
0x8c - (0x8b - 0x85),	/* M-F12 */
0x8d - (0x8d - 0x48),	/* C-Up */
0x8e,
0x8f,
0x90,
0x91 - (0x91 - 0x50),	/* C-Down */
0x92 - (0x92 - 0x52),	/* C-Insert */
0x93 - (0x93 - 0x53),	/* C-Delete */
0x94,
0x95,
0x96,
0x97,
0x98,
0x99,
0x9a,
0x9b,
0x9c,
0x9d,
0x9e,
0x9f,
0xa0,
0xa1,
0xa2,
0xa3,
0xa4,
0xa5,
0xa6,
0xa7,
0xa8,
0xa9,
0xaa,
0xab,
0xac,
0xad,
0xae,
0xaf,
0xb0,
0xb1,
0xb2,
0xb3,
0xb4,
0xb5,
0xb6,
0xb7,
0xb8,
0xb9,
0xba,
0xbb,
0xbc,
0xbd,
0xbe,
0xbf,
0xc0,
0xc1,
0xc2,
0xc3,
0xc4,
0xc5,
0xc6,
0xc7,
0xc8,
0xc9,
0xca,
0xcb,
0xcc,
0xcd,
0xce,
0xcf,
0xd0,
0xd1,
0xd2,
0xd3,
0xd4,
0xd5,
0xd6,
0xd7,
0xd8,
0xd9,
0xda,
0xdb,
0xdc,
0xdd,
0xde,
0xdf,
0xe0,
0xe1,
0xe2,
0xe3,
0xe4,
0xe5,
0xe6,
0xe7,
0xe8,
0xe9,
0xea,
0xeb,
0xec,
0xed,
0xee,
0xef,
0xf0,
0xf1,
0xf2,
0xf3,
0xf4,
0xf5,
0xf6,
0xf7,
0xf8,
0xf9,
0xfa,
0xfb,
0xfc,
0xfd,
0xfe,
0xff,
};

static KeyCode _WGetKeycode (_WQueueEvent *ev)
{
  int scan = ((ev->scan >> 8) & 0xff);
  return (scancode_remap[scan] + 8);
}

_XReadEvents (dpy)
     register Display *dpy;
{
  _WQueueEvent ev;
  XEvent xe;
  Window pointer_window, enter_window, leave_window, root, window;
  unsigned int state;
  int button;
  int i;

  /* read device events and break only of dpy->qlen > 0 */

  while (1) {

    _WQueueGetEvent (M_EVENT|M_NOPAINT, &ev);

    xe.xany.send_event = 0;
    xe.xany.display = NULL;

    root = DefaultRootWindow (dpy);
    pointer_window = _WPointToWindow (dpy, ev.x, ev.y);
    state = _WButtonAndModifierState (&ev);
    button = _WEventButton (&ev);

    if ((ev.flags & M_KEYPRESS) && _Focus.window != None) {

      window = (_Focus.window != (Window) PointerRoot) 
	? _Focus.window
	: pointer_window;

      xe.xkey.type = KeyPress;
      xe.xkey.window = window;
      xe.xkey.root = root;
      xe.xkey.subwindow = (pointer_window->parent == window) ? pointer_window : None;
      xe.xkey.time = _WConvertTime (ev.time);
      xe.xkey.x = ev.x - window->window_port.x1;
      xe.xkey.y = ev.y - window->window_port.y1;
      xe.xkey.x_root = ev.x;
      xe.xkey.y_root = ev.y;
      xe.xkey.state = state;
      xe.xkey.keycode = _WGetKeycode (&ev);
      xe.xkey.same_screen = True;
      _WPropagateEvent (&xe, window);
    }

    if (ev.flags & M_BUTTON_DOWN) {

      window = None;
      if (_PointerGrab.active) {
	if (_PointerGrab.owner_events == False) {
	  if (_PointerGrab.event_mask & ButtonPressMask) {
	    window = _PointerGrab.grab_window;
	    xe.xbutton.type = ButtonPress;
	    xe.xbutton.display = _PointerGrab.display;
	    xe.xbutton.window = window;
	    xe.xbutton.root = root;
	    xe.xbutton.subwindow = None;
	    xe.xbutton.time = _WConvertTime (ev.time);
	    xe.xbutton.x = ev.x - window->window_port.x1;
	    xe.xbutton.y = ev.y - window->window_port.y1;
	    xe.xbutton.x_root = ev.x;
	    xe.xbutton.y_root = ev.y;
	    /* do not include button into state */
	    xe.xbutton.state = state & ~_WButtonToMask (button);
	    xe.xbutton.button = button;
	    xe.xbutton.same_screen = True;
	    _XEnq (xe.xbutton.display, &xe);
	  }
	}
	else {
	  window = _WFindEventWindow (_PointerGrab.display, pointer_window,
				      ButtonPressMask);
	  if (window == None && (_PointerGrab.event_mask & ButtonPressMask))
	    window = _PointerGrab.grab_window;
	  if (window != None) {
	    xe.xbutton.type = ButtonPress;
	    xe.xbutton.display = _PointerGrab.display;
	    xe.xbutton.window = window;
	    xe.xbutton.root = root;
	    xe.xbutton.subwindow = None;
	    xe.xbutton.time = _WConvertTime (ev.time);
	    xe.xbutton.x = ev.x - window->window_port.x1;
	    xe.xbutton.y = ev.y - window->window_port.y1;
	    xe.xbutton.x_root = ev.x;
	    xe.xbutton.y_root = ev.y;
	    /* do not include button into state */
	    xe.xbutton.state = state & ~_WButtonToMask (button);
	    xe.xbutton.button = button;
	    xe.xbutton.same_screen = True;
	    _XEnq (xe.xbutton.display, &xe);
	  }
	}
      }
      else {	/* pointer grab not active */
	xe.xbutton.type = ButtonPress;
	xe.xbutton.root = root;
	xe.xbutton.subwindow = None;
	xe.xbutton.time = _WConvertTime (ev.time);
	xe.xbutton.x_root = ev.x;
	xe.xbutton.y_root = ev.y;
	/* do not include button into state */
	xe.xbutton.state = state & ~_WButtonToMask (button);
	xe.xbutton.button = button;
	xe.xbutton.same_screen = True;
	_WPropagateEvent (&xe, pointer_window);

	/* activate an internal pointer grab */

	window = pointer_window;
	for (i = 0; i < MAX_CONN; i++) {
	  if (window->event_masks[i] & ButtonPressMask)
	    break;
	}
	if (i < MAX_CONN) {
	  _PointerGrab.display = _ConnTable[i];
	  _PointerGrab.grab_window = window;
	  _PointerGrab.owner_events = 
	    (window->event_masks[i] & OwnerGrabButtonMask) ? True : False;
	  _PointerGrab.event_mask = window->event_masks[i];
	  _PointerGrab.pointer_mode = GrabModeAsync;
	  _PointerGrab.keyboard_mode = GrabModeAsync;
	  _PointerGrab.confine_to = None;
	  _PointerGrab.cursor = None;
	  _PointerGrab.time = _WCurrentTime ();
	  _PointerGrab.active = 2;
	}
      }
    }

    if (ev.flags & M_BUTTON_UP) {

      window = None;
      if (_PointerGrab.active) {
	if (_PointerGrab.owner_events == False) {
	  if (_PointerGrab.event_mask & ButtonReleaseMask) {
	    window = _PointerGrab.grab_window;
	    xe.xbutton.type = ButtonRelease;
	    xe.xbutton.display = _PointerGrab.display;
	    xe.xbutton.window = window;
	    xe.xbutton.root = root;
	    xe.xbutton.subwindow = None;
	    xe.xbutton.time = _WConvertTime (ev.time);
	    xe.xbutton.x = ev.x - window->window_port.x1;
	    xe.xbutton.y = ev.y - window->window_port.y1;
	    xe.xbutton.x_root = ev.x;
	    xe.xbutton.y_root = ev.y;
	    xe.xbutton.state = state | _WButtonToMask (button);
	    xe.xbutton.button = button;
	    xe.xbutton.same_screen = True;
	    _XEnq (xe.xbutton.display, &xe);
	  }
	}
	else {
	  window = _WFindEventWindow (_PointerGrab.display, pointer_window,
				      ButtonReleaseMask);
	  if (window == None && (_PointerGrab.event_mask & ButtonReleaseMask))
	    window = _PointerGrab.grab_window;
	  if (window != None) {
	    xe.xbutton.type = ButtonRelease;
	    xe.xbutton.display = _PointerGrab.display;
	    xe.xbutton.window = window;
	    xe.xbutton.root = root;
	    xe.xbutton.subwindow = None;
	    xe.xbutton.time = _WConvertTime (ev.time);
	    xe.xbutton.x = ev.x - window->window_port.x1;
	    xe.xbutton.y = ev.y - window->window_port.y1;
	    xe.xbutton.x_root = ev.x;
	    xe.xbutton.y_root = ev.y;
	    xe.xbutton.state = state | _WButtonToMask (button);
	    xe.xbutton.button = button;
	    xe.xbutton.same_screen = True;
	    _XEnq (xe.xbutton.display, &xe);
	  }
	}
	if (_PointerGrab.active == 2 && ev.buttons == 0) {
	  _PointerGrab.active = 0;
	  _PointerGrab.display = NULL;
	}
      }
      else {
	xe.xbutton.type = ButtonRelease;
	xe.xbutton.root = root;
	xe.xbutton.subwindow = None;
	xe.xbutton.time = _WConvertTime (ev.time);
	xe.xbutton.x_root = ev.x;
	xe.xbutton.y_root = ev.y;
	xe.xbutton.state = state | _WButtonToMask (button);
	xe.xbutton.button = button;
	xe.xbutton.same_screen = True;
	_WPropagateEvent (&xe, pointer_window);
      }
    }

    if (ev.flags & M_MOTION) {
      window = None;
      if (_PointerGrab.active) {
	if (_PointerGrab.owner_events == False) {
	  if (_PointerGrab.event_mask & AllMotionMask) {
	    window = _PointerGrab.grab_window;
	    xe.xmotion.type = MotionNotify;
	    xe.xmotion.display = _PointerGrab.display;
	    xe.xmotion.window = window;
	    xe.xmotion.root = root;
	    xe.xmotion.subwindow = None;
	    xe.xmotion.time = _WConvertTime (ev.time);
	    xe.xmotion.x = ev.x - window->window_port.x1;
	    xe.xmotion.y = ev.y - window->window_port.y1;
	    xe.xmotion.x_root = ev.x;
	    xe.xmotion.y_root = ev.y;
	    xe.xmotion.state = state;
	    xe.xmotion.is_hint = False;
	    xe.xmotion.same_screen = True;
	    _XEnq (xe.xmotion.display, &xe);
	  }
	}
	else {
	  window = _WFindEventWindow (_PointerGrab.display, pointer_window,
				      AllMotionMask);
	  if (window == None && (_PointerGrab.event_mask & AllMotionMask))
	    window = _PointerGrab.grab_window;
	  if (window != None) {
	    xe.xmotion.type = MotionNotify;
	    xe.xmotion.display = _PointerGrab.display;
	    xe.xmotion.window = window;
	    xe.xmotion.root = root;
	    xe.xmotion.subwindow = (pointer_window->parent == window)
	      ? pointer_window : None;
	    xe.xmotion.time = _WConvertTime (ev.time);
	    xe.xmotion.x = ev.x - window->window_port.x1;
	    xe.xmotion.y = ev.y - window->window_port.y1;
	    xe.xmotion.x_root = ev.x;
	    xe.xmotion.y_root = ev.y;
	    xe.xmotion.state = state;
	    xe.xmotion.is_hint = False;
	    xe.xmotion.same_screen = True;
	    _XEnq (xe.xmotion.display, &xe);
	  }
	}
      }
      /* check for enter/leave window */

      if (_LastMouseWindow == None)
	_LastMouseWindow = root;

      enter_window = pointer_window;
      leave_window = _LastMouseWindow;

      if (enter_window != leave_window) {

	if (_PointerGrab.active == 0 || _PointerGrab.cursor == None)
	  _WDefineCursor (_WGetWindowCursor (pointer_window));

	if (leave_window->event_mask & LeaveWindowMask) {
	  xe.xcrossing.type = LeaveNotify;
	  xe.xcrossing.window = leave_window;
	  xe.xcrossing.root = root;
	  xe.xcrossing.subwindow = None;
	  xe.xcrossing.time = _WConvertTime (ev.time);
	  xe.xcrossing.x = ev.x - leave_window->window_port.x1;
	  xe.xcrossing.y = ev.y - leave_window->window_port.y1;
	  xe.xcrossing.x_root = ev.x;
	  xe.xcrossing.y_root = ev.y;
	  xe.xcrossing.mode = NotifyNormal;
	  if (_WIsInferior (leave_window, enter_window))
	    xe.xcrossing.detail = NotifyAncestor;
	  else if (_WIsInferior (enter_window, leave_window))
	    xe.xcrossing.detail = NotifyInferior;
	  else
	    xe.xcrossing.detail = NotifyNonlinear;
	  xe.xcrossing.same_screen = True;
	  xe.xcrossing.focus = _WIsFocus (root, xe.xcrossing.window);
	  xe.xcrossing.state = state;
	  _WDispatchEvent (&xe);
	}
	if (enter_window->event_mask & EnterWindowMask) {
	  xe.xcrossing.type = EnterNotify;
	  xe.xcrossing.window = enter_window;
	  xe.xcrossing.root = root;
	  xe.xcrossing.subwindow = None;
	  xe.xcrossing.time = _WConvertTime (ev.time);
	  xe.xcrossing.x = ev.x - enter_window->window_port.x1;
	  xe.xcrossing.y = ev.y - enter_window->window_port.y1;
	  xe.xcrossing.x_root = ev.x;
	  xe.xcrossing.y_root = ev.y;
	  xe.xcrossing.mode = NotifyNormal;
	  if (_WIsInferior (enter_window, leave_window))
	    xe.xcrossing.detail = NotifyAncestor;
	  else if (_WIsInferior (leave_window, enter_window))
	    xe.xcrossing.detail = NotifyInferior;
	  else
	    xe.xcrossing.detail = NotifyNonlinear;
	  xe.xcrossing.same_screen = True;
	  xe.xcrossing.focus = _WIsFocus (root, xe.xcrossing.window);
	  xe.xcrossing.state = state;
	  _WDispatchEvent (&xe);
	}
      }
      else if (! _PointerGrab.active) {
	xe.xmotion.type = MotionNotify;
	xe.xmotion.root = root;
	xe.xmotion.subwindow = None;
	xe.xmotion.time = _WConvertTime (ev.time);
	xe.xmotion.x_root = ev.x;
	xe.xmotion.y_root = ev.y;
	xe.xmotion.state = state;
	xe.xmotion.is_hint = False;
	xe.xmotion.same_screen = True;
	_WPropagateEvent (&xe, pointer_window);
      }
    }
    _LastMouseEvent = ev;
    _LastMouseWindow = pointer_window;

    if (dpy->qlen > 0 || _DoNotBlock == True) break;
  }
}

/*
 * _XEnq - Place event packets on the display's queue.
 */
void
_XEnq (Display *dpy, XEvent *event)
{
	register _XQEvent *qelt;

	event->xany.serial = ++dpy->request;
/*NOSTRICT*/
	if (qelt = _qfree) {
		/* If _qfree is non-NULL do this, else malloc a new one. */
		_qfree = qelt->next;
	}
	else if ((qelt = 
	    (_XQEvent *) Xmalloc((unsigned)sizeof(_XQEvent))) == NULL) {
		/* Malloc call failed! */
		errno = ENOMEM;
		_XIOError(dpy);
	}
	qelt->next = NULL;
	qelt->event = *event;
	if (dpy->tail)	dpy->tail->next = qelt;
	else 		dpy->head = qelt;
	dpy->tail = qelt;
	dpy->qlen++;
}
