/* $Id: eventque.c 1.1 1994/01/26 19:02:45 ulrich Exp $ */
/*
 * eventque.c
 *
 * Event Queue functions, source code partly from GRX 1.03 mouse.c.
 */
#include "Xlibemu.h"

#include <grx.h>
#include <mousex.h>
#include <../events/eventque.h>

/* from contrib/libgrx/src/libgrx.h */
extern  char _GrMouseDrawn;		/* set if mouse drawn */
extern  char _GrMouseCheck;		/* set if mouse blocking needed */

extern  int  (*_GrMouseBlock)(GrContext *c,int x1,int y1,int x2,int y2);
extern  void (*_GrMouseUnBlock)(int);
extern  void (*_GrMouseUnInit)(void);

extern  GrContext _GrContext;		/* current graphics context */
extern  GrContext _GrVidPage;		/* the whole screen */

#define CURC		(&_GrContext)
#define SCREEN		(&_GrVidPage)


#include <string.h>
#include <stdarg.h>

#ifdef __TURBOC__
# include <time.h>
# include <conio.h>
#endif

#ifdef __GNUC__
  extern  long	    rawclock();
# define  clock()   rawclock()
#endif

/*
 * queue status flags
 */
#define UNKNOWN	    0
#define INITTED	    1

#define DRAW_STACK  1024
#define QUEUE_SIZE  100

#define CHECK_CONTEXT(cxt)  ((cxt)->gc_baseaddr == SCREEN->gc_baseaddr)

extern GrCursor *_GrMouseCursor;	/* current mouse cursor */

static EventQueue *queue;
static int queue_status = UNKNOWN;
static int own_cursor	= FALSE;
static int kb_enable	= TRUE;
static int ms_enable	= TRUE;
static int ms_buttons;

static char ptr12x16bits[] = {
    0,1,0,0,0,0,0,0,0,0,0,0,
    1,2,1,0,0,0,0,0,0,0,0,0,
    1,2,2,1,0,0,0,0,0,0,0,0,
    1,2,2,2,1,0,0,0,0,0,0,0,
    1,2,2,2,2,1,0,0,0,0,0,0,
    1,2,2,2,2,2,1,0,0,0,0,0,
    1,2,2,2,2,2,2,1,0,0,0,0,
    1,2,2,2,2,2,2,2,1,0,0,0,
    1,2,2,2,2,2,2,2,2,1,0,0,
    1,2,2,2,2,2,2,2,2,2,1,0,
    1,2,2,2,2,2,2,2,2,2,2,1,
    1,2,2,2,2,1,1,1,1,1,1,0,
    1,2,2,2,1,0,0,0,0,0,0,0,
    1,2,2,1,0,0,0,0,0,0,0,0,
    1,2,1,0,0,0,0,0,0,0,0,0,
    0,1,0,0,0,0,0,0,0,0,0,0,
};


static void draw_mouse(void)
{
  if(_GrMouseCursor->cr_displayed) return;
  GrMoveCursor(_GrMouseCursor,queue->evq_xpos,queue->evq_ypos);
  GrDisplayCursor(_GrMouseCursor);
}

static void erase_mouse(void)
{
  if(!_GrMouseCursor->cr_displayed) return;
  GrEraseCursor(_GrMouseCursor);
}

static void move_mouse(void)
{
  GrMoveCursor(_GrMouseCursor,queue->evq_xpos,queue->evq_ypos);
}

static void get_queued_event(int flags,_WQueueEvent *e)
{
	EventRecord evr;
	int moved = 0;

	for( ; ; ) {
	    if(queue->evq_moved) {
		moved = M_MOTION;
		queue->evq_moved = 0;
	    }
	    if(EventQueueNextEvent(queue,&evr)) {
		switch(evr.evt_type) {
		  case EVENT_KEYBD:
		    if(flags & M_KEYPRESS) {
			e->flags   = M_KEYPRESS;
			e->x	   = queue->evq_xpos;
			e->y	   = queue->evq_ypos;
			e->key	   = evr.evt_keycode;
			e->scan    = evr.evt_scancode;
			e->kbstat  = evr.evt_kbstat;
			e->time	   = evr.evt_time;
			return;
		    }
		    break;
		  case EVENT_MOUSE:
		    ms_buttons = evr.evt_button;
		    if(flags & evr.evt_mask) {
			e->flags   = flags & (evr.evt_mask | moved);
			e->x	   = evr.evt_xpos;
			e->y	   = evr.evt_ypos;
			e->buttons = evr.evt_button;
			e->kbstat  = evr.evt_kbstat;
			e->time	   = evr.evt_time;
			return;
		    }
		    break;
		}
		continue;
	    }
	    if(flags & (moved | M_POLL)) {
		e->flags   = flags & moved;
		e->x	   = queue->evq_xpos;
		e->y	   = queue->evq_ypos;
		e->buttons = ms_buttons;
		e->kbstat  = getkbstat();
		e->time	   = clock();
		return;
	    }
	}
}


void _WQueueUnInit(void)
{
  if(queue_status) {
    EventQueueDeInit();
    queue = NULL;
  }
  if(_GrMouseDrawn) erase_mouse();
  queue_status = UNKNOWN;
}

void _WQueueInit(void)
{
  if(queue_status) {
    _WQueueUnInit();  
  }
  queue = EventQueueInit(QUEUE_SIZE,DRAW_STACK,move_mouse);
  if(queue != NULL) {
    queue_status = INITTED;
    queue->evq_drawmouse = FALSE;
  }
  if(_GrMouseCursor == NULL) _WMouseSetColors(GrWhite(),GrBlack());
  _GrMouseBlock	= _WMouseBlock;
  _GrMouseUnBlock = _WMouseUnBlock;
  _GrMouseUnInit  = _WQueueUnInit;
  _GrMouseDrawn	= FALSE;
  _GrMouseCheck	= FALSE;

  _WMouseEventEnable(kb_enable,ms_enable);
  _WMouseSetLimits(0,0,GrScreenX() - 1,GrScreenY() - 1);
  _WMouseWarp((GrScreenX() >> 1),(GrScreenY() >> 1));
  _WMouseSetSpeed(1);
  _WMouseSetAccel(1000,1);
}

void _WMouseSetSpeed(int speed)
{
  if(!queue_status) return;
  queue->evq_xspeed = speed;
  queue->evq_yspeed = speed;
}

void _WMouseSetAccel(int thresh,int accel)
{
  if(!queue_status) return;
  queue->evq_thresh = thresh;
  queue->evq_accel  = accel;
}

void _WMouseSetLimits(int x1,int y1,int x2,int y2)
{
  if(!queue_status) return;
  if(x1 < 0) x1 = 0;
  if(y1 < 0) y1 = 0;
  if(x2 >= GrScreenX()) x2 = GrScreenX()-1;
  if(y2 >= GrScreenX()) y2 = GrScreenY()-1;
  queue->evq_xmin = x1;
  queue->evq_ymin = y1;
  queue->evq_xmax = x2;
  queue->evq_ymax = y2;
  _WMouseWarp(queue->evq_xpos,queue->evq_ypos);
}

void _WMouseGetLimits(int *x1,int *y1,int *x2,int *y2)
{
  if(!queue_status) return;
  *x1 = queue->evq_xmin;
  *y1 = queue->evq_ymin;
  *x2 = queue->evq_xmax;
  *y2 = queue->evq_ymax;
}

void _WMouseWarp(int x,int y)
{
  char msdraw;

  if(!queue_status) return;
  msdraw = queue->evq_drawmouse;
  queue->evq_drawmouse = FALSE;
  if(x < queue->evq_xmin) x = queue->evq_xmin;
  if(y < queue->evq_ymin) y = queue->evq_ymin;
  if(x > queue->evq_xmax) x = queue->evq_xmax;
  if(y > queue->evq_ymax) y = queue->evq_ymax;
  queue->evq_xpos = x;
  queue->evq_ypos = y;
  if(_GrMouseDrawn) move_mouse();
  queue->evq_drawmouse = msdraw;
}

void _WMouseSetCursor(GrCursor *cursor)
{
  GrCursor *old = _GrMouseCursor;
  char msdraw = 0;

  if(cursor != NULL) {
    if(queue_status) {
      msdraw = queue->evq_drawmouse;
      queue->evq_drawmouse = FALSE;
      if(_GrMouseDrawn) erase_mouse();
    }
    _GrMouseCursor = cursor;
    if(own_cursor) GrDestroyCursor(old);
    own_cursor = FALSE;
    if(queue_status) {
      if(_GrMouseDrawn) draw_mouse();
      queue->evq_drawmouse = msdraw;
    }
  }
}

void _WMouseSetColors(int fg,int bg)
{
  GrCursor *new;
  int cols[3];
  
  cols[0] = 2;
  cols[1] = bg;
  cols[2] = fg;
  new = GrBuildCursor(ptr12x16bits,12,16,1,1,cols);
  if(new != NULL) {
    _WMouseSetCursor(new);
    own_cursor = TRUE;
  }
}

void _WMouseDisplayCursor(void)
{
  if(!queue_status || _GrMouseDrawn) return;
  draw_mouse();
  _GrMouseDrawn = TRUE;
  _GrMouseCheck = CHECK_CONTEXT(CURC) ? TRUE : FALSE;
  queue->evq_drawmouse = TRUE;
}

void _WMouseEraseCursor(void)
{
  if(!queue_status || !_GrMouseDrawn) return;
  queue->evq_drawmouse = FALSE;
  _GrMouseDrawn = FALSE;
  _GrMouseCheck = FALSE;
  erase_mouse();
}

#define BLK_MOTION  1		/* "MouseBlock" froze the mouse */
#define BLK_DISPLAY 2		/* "MouseBlock" erased the mouse */
#define BLK_CHECK   4		/* "MouseBlock" cleared check flag */

int _WMouseBlock(GrContext *c,int x1,int y1,int x2,int y2)
{
  int mx1,my1,mx2,my2;
  int retval;

  if(!queue_status) return(0);
  if(!_GrMouseDrawn) return(0);
  if(c == NULL) c = CURC;
  if(!CHECK_CONTEXT(c)) return(0);
  retval = queue->evq_drawmouse ? BLK_MOTION : 0;
  queue->evq_drawmouse = FALSE;
  queue->evq_moved = FALSE;
  if (x1 > x2) { int tmp = x1; x1 = x2; x2 = tmp; }
  if (y1 > y2) { int tmp = y1; y1 = y2; y2 = tmp; }
  x1 += c->gc_xoffset; y1 += c->gc_yoffset;
  x2 += c->gc_xoffset; y2 += c->gc_yoffset;
  mx1 = _GrMouseCursor->cr_xwpos;
  my1 = _GrMouseCursor->cr_ywpos;
  mx2 = mx1 + _GrMouseCursor->cr_xwork - 1;
  my2 = my1 + _GrMouseCursor->cr_ywork - 1;
  if(mx1 > x1) x1 = mx1;
  if(my1 > y1) y1 = my1;
  if(mx2 < x2) x2 = mx2;
  if(my2 < y2) y2 = my2;
  if((x1 <= x2) && (y1 <= y2)) {
    _GrMouseDrawn = FALSE;
    retval |= BLK_DISPLAY;
    erase_mouse();
  }
  if(_GrMouseCheck) {
    _GrMouseCheck = FALSE;
    retval |= BLK_CHECK;
  }
  return(retval);
}

void _WMouseUnBlock(int flags)
{
  if(!queue_status) return;
  if(queue->evq_moved && (flags & BLK_MOTION)) move_mouse();
  if(!_GrMouseDrawn && (flags & BLK_DISPLAY)) {
    draw_mouse();
    _GrMouseDrawn = TRUE;
  }
  _GrMouseCheck = ((CHECK_CONTEXT(CURC) && (flags & BLK_CHECK)) ?
		   TRUE :
		   FALSE
		   );
  if(flags & BLK_MOTION) queue->evq_drawmouse = TRUE;
}

void _WMouseEventEnable(int enable_kb,int enable_ms)
{
  if(queue_status) {
    queue->evq_enable =
      (enable_kb ? EVENT_ENABLE(EVENT_KEYBD) : 0) |
	(enable_ms ? EVENT_ENABLE(EVENT_MOUSE) : 0);
  }
  kb_enable = enable_kb;
  ms_enable = enable_ms;
}

void _WQueueGetEvent(int flags,_WQueueEvent *e)
{
  int erase = FALSE;

  if(!kb_enable) flags &= ~M_KEYPRESS;
  if(!ms_enable) flags &= ~(M_MOTION | M_BUTTON_DOWN | M_BUTTON_UP);
  if(!(flags & M_EVENT)) { e->flags = 0; return; }
  if(!queue_status) {
    e->flags = 0;
    return;
  }
  if(!_GrMouseDrawn && !(flags & M_NOPAINT)) {
    draw_mouse();
    queue->evq_drawmouse = TRUE;
    erase = TRUE;
  }
  get_queued_event (flags,e);
  if(erase) {
    queue->evq_drawmouse = FALSE;
    erase_mouse();
  }
}

int _WQueuePendingEvent(void)
{
  if(queue_status
     && ((queue->evq_cursize > 0) || (queue->evq_moved != 0)))
    return(TRUE);
  return(FALSE);
}

void _WQueueQuery (_WQueueEvent *e)
{
  e->flags	= 0;
  e->x		= queue->evq_xpos;
  e->y		= queue->evq_ypos;
  e->buttons	= ms_buttons;
  e->key	= 0;
  e->scan	= 0;
  e->kbstat	= getkbstat();
  e->time	= clock();
}
