/* $Id: gc.c 1.3 1994/01/26 19:14:46 ulrich Exp $ */
/*
 * gc.c
 *
 * X library GC functions.
 */

#include "Xlibemu.h"

void
GxPatternedLine (int x1, int y1, int x2, int y2, GrLinePattern *lnp)
{
  GrPatternFilledLine (x1, y1, x2, y2, lnp->lnp_pattern);
}

void
GxPatternedBox (int x1, int y1, int x2, int y2, GrLinePattern *lnp)
{
  GrPatternFilledLine (x1, y1, x2, y1, lnp->lnp_pattern);
  if (y1 == y2) return;
  GrPatternFilledLine (x1, y2, x2, y2, lnp->lnp_pattern);
  GrPatternFilledLine (x1, y1+1, x1, y2-1, lnp->lnp_pattern);
  GrPatternFilledLine (x2, y1+1, x2, y2-1, lnp->lnp_pattern);
}

const unsigned long  _GXfunctionToGrOPER[16] = {
  GrWRITE ,			/* GXclear		0x0 0 */
  GrAND   ,			/* GXand		0x1 src AND dst */
  GrAND   ,			/* GXandReverse		0x2 src AND NOT dst */
  GrWRITE ,			/* GXcopy		0x3 src */
  GrAND   ,			/* GXandInverted	0x4 NOT src AND dst */
  GrWRITE ,			/* GXnoop		0x5 dst */
  GrXOR   ,			/* GXxor		0x6 src XOR dst */
  GrOR    ,			/* GXor			0x7 src OR dst */
  GrAND   ,			/* GXnor		0x8 NOT src AND NOT dst */
  GrXOR   ,			/* GXequiv		0x9 NOT src XOR dst */
  GrXOR|0xffff ,		/* GXinvert		0xa NOT dst */
  GrOR    ,			/* GXorReverse		0xb src OR NOT dst */
  GrWRITE ,			/* GXcopyInverted	0xc NOT src */
  GrOR    ,			/* GXorInverted		0xd NOT src OR dst */
  GrOR    ,			/* GXnand		0xe NOT src OR NOT dst */
  GrWRITE ,			/* GXset		0xf 1 */
};

static const unsigned long  map_function_src_oper[16][2] =
{
{ GrAND,	0x0000 }, /* GXclear		0x0 0 */
{ GrAND,	0xffff }, /* GXand		0x1 src AND dst */
{ GrAND,	0xffff }, /* GXandReverse	0x2 src AND NOT dst */
{ GrWRITE,	0xffff }, /* GXcopy		0x3 src */
{ GrXOR,	0xffff }, /* GXandInverted	0x4 NOT src AND dst */
{ GrAND,	0x0000 }, /* GXnoop		0x5 dst */
{ GrXOR,	0xffff }, /* GXxor		0x6 src XOR dst */
{ GrAND,	0xffff }, /* GXor		0x7 src OR dst */
{ GrXOR,	0xffff }, /* GXnor		0x8 NOT src AND NOT dst */
{ GrXOR,	0xffff }, /* GXequiv		0x9 NOT src XOR dst */
{ GrAND,	0x0000 }, /* GXinvert		0xa NOT dst */
{ GrAND,	0xffff }, /* GXorReverse	0xb src OR NOT dst */
{ GrXOR,	0xffff }, /* GXcopyInverted	0xc NOT src */
{ GrXOR,	0xffff }, /* GXorInverted	0xd NOT src OR dst */
{ GrXOR,	0xffff }, /* GXnand		0xe NOT src OR NOT dst */
{ GrAND,	0x0000 }};/* GXset		0xf 1 */

static const unsigned long  map_function_dst_oper[16][2] = {
{ GrAND,	0x0000 }, /* GXclear		0x0 0 */
{ GrAND,	0xffff }, /* GXand		0x1 src AND dst */
{ GrXOR,	0xffff }, /* GXandReverse	0x2 src AND NOT dst */
{ GrAND,	0x0000 }, /* GXcopy		0x3 src */
{ GrAND,	0xffff }, /* GXandInverted	0x4 NOT src AND dst */
{ GrAND,	0xffff }, /* GXnoop		0x5 dst */
{ GrAND,	0xffff }, /* GXxor		0x6 src XOR dst */
{ GrAND,	0xffff }, /* GXor		0x7 src OR dst */
{ GrXOR,	0xffff }, /* GXnor		0x8 NOT src AND NOT dst */
{ GrAND,	0xffff }, /* GXequiv		0x9 NOT src XOR dst */
{ GrXOR,	0xffff }, /* GXinvert	0xa NOT dst */
{ GrXOR,	0xffff }, /* GXorReverse	0xb src OR NOT dst */
{ GrAND,	0x0000 }, /* GXcopyInverted	0xc NOT src */
{ GrAND,	0xffff }, /* GXorInverted	0xd NOT src OR dst */
{ GrXOR,	0xffff }, /* GXnand		0xe NOT src OR NOT dst */
{ GrAND,	0x0000 }};/* GXset		0xf 1 */


static const _WGCExt initial_Ext =
{   GrLine,
    GrBox,
    GrCircle,
    GrEllipse,
    GrCircleArc,
    GrEllipseArc,
    GrPolyLine,
    GrPolygon,
    GrFilledBox,
    GrFilledCircle,
    GrFilledEllipse,
    GrFilledCircleArc,
    GrFilledEllipseArc,
    GrFilledPolygon,
      0l,		/* line_arg */
      0l,		/* fill_arg */
    };


static XGCValues const initial_GC = {
    GXcopy, 	/* function */
    AllPlanes,	/* plane_mask */
    0L,		/* foreground */
    1L,		/* background */
    0,		/* line_width */
    LineSolid,	/* line_style */
    CapButt,	/* cap_style */
    JoinMiter,	/* join_style */
    FillSolid,	/* fill_style */
    EvenOddRule,/* fill_rule */
    ArcPieSlice,/* arc_mode */
    (Pixmap)~0L,/* tile, impossible (unknown) resource */
    (Pixmap)~0L,/* stipple, impossible (unknown) resource */
    0,		/* ts_x_origin */
    0,		/* ts_y_origin */
    (Font)~0L,	/* font, impossible (unknown) resource */
    ClipByChildren, /* subwindow_mode */
    True,	/* graphics_exposures */
    0,		/* clip_x_origin */
    0,		/* clip_y_origin */
    None,	/* clip_mask */
    0,		/* dash_offset */
    4		/* dashes (list [4,4]) */
};

void
_WUpdateGC (GC gc)
{
  int oper;
  GrPattern *pat;
  _WGCExt *ext = (_WGCExt *)gc->ext_data;

  oper = _GXfunctionToGrOPER[gc->values.function];

  switch (gc->values.fill_style) {
  case FillSolid:
    ext->fill_arg = oper | gc->values.foreground;
    break;
  case FillTiled:
    ext->fill_arg = (long) pat = gc->values.tile->pattern;
    if (gc->values.tile->type == 1) {
      ((GrPattern *) gc->values.tile)->gp_pxp_oper =oper;
    }  
    break;
  case FillStippled:
    ext->fill_arg = (long) pat = gc->values.stipple->pattern;
    if (gc->values.stipple->type == 0) {
      pat->gp_bmp_fgcolor = oper | gc->values.foreground;
      pat->gp_bmp_bgcolor = GrNOCOLOR;
    }
    break;
  case FillOpaqueStippled:
    ext->fill_arg = (long) pat = gc->values.stipple->pattern;
    if (gc->values.stipple->type == 0) {
      pat->gp_bmp_fgcolor = oper | gc->values.foreground;
      pat->gp_bmp_bgcolor = oper | gc->values.background;
    }
    break;
  }

  if (gc->values.fill_style == FillSolid &&
      gc->values.line_width <= 1) {
    ext->line_arg = ext->fill_arg;

    ext->Line	= GrLine;
    ext->Box	= GrBox;
    ext->Circle	= GrCircle;
    ext->Ellipse	= GrEllipse;
    ext->CircleArc	= GrCircleArc;
    ext->EllipseArc	= GrEllipseArc;
    ext->PolyLine	= GrPolyLine;
    ext->Polygon	= GrPolygon;
  }
  else {
    ext->line_opt.lno_width = gc->values.line_width;
    ext->line_opt.lno_dashpat = 0;	/* no dashes for now */
    ext->line_opt.lno_pattlen = 0;
    if (gc->values.fill_style == FillSolid) {
      ext->line_opt.lno_color = ext->fill_arg;
      ext->line_arg = (long) &ext->line_opt;

      ext->Line	= GrCustomLine;
      ext->Box	= GrCustomBox;
      ext->Circle	= GrCustomCircle;
      ext->Ellipse	= GrCustomEllipse;
      ext->CircleArc	= GrCustomCircleArc;
      ext->EllipseArc	= GrCustomEllipseArc;
      ext->PolyLine	= GrCustomPolyLine;
      ext->Polygon	= GrCustomPolygon;
    }
    else {
      ext->line_pat.lnp_pattern = (GrPattern *) ext->fill_arg;
      ext->line_pat.lnp_option = &ext->line_opt;
      ext->line_arg = (long) &ext->line_pat;

      ext->Line	= GxPatternedLine;
      ext->Box	= GxPatternedBox;
      ext->Circle	= GrPatternedCircle;
      ext->Ellipse	= GrPatternedEllipse;
      ext->CircleArc	= GrPatternedCircleArc;
      ext->EllipseArc	= GrPatternedEllipseArc;
      ext->PolyLine	= GrPatternedPolyLine;
      ext->Polygon	= GrPatternedPolygon;
    }
  }

  if (gc->values.fill_style == FillSolid) {
    ext->FilledBox	= GrFilledBox;
    ext->FilledCircle	= GrFilledCircle;
    ext->FilledEllipse	= GrFilledEllipse;
    ext->FilledCircleArc	= GrFilledCircleArc;
    ext->FilledEllipseArc	= GrFilledEllipseArc;
    ext->FilledPolygon	= GrFilledPolygon; 
  }
  else {
    ext->FilledBox	= GrPatternFilledBox;
    ext->FilledCircle	= GrPatternFilledCircle;
    ext->FilledEllipse	= GrPatternFilledEllipse;
    ext->FilledCircleArc	= GrPatternFilledCircleArc;
    ext->FilledEllipseArc	= GrPatternFilledEllipseArc;
    ext->FilledPolygon	= GrPatternFilledPolygon; 
  }
  if (gc->values.font == (Font)~0L) {
    GrTextOption opt;
    if (GrFindBestFont (8, 16, 0, "X_misc*", &opt))
      gc->values.font = opt.txo_font;
    else if (GrFindBestFont (8, 16, 0, "pc", &opt))
      gc->values.font = opt.txo_font;
    else
      gc->values.font = (Font)0;
  }
}


GC XCreateGC(
    Display*		display,
    Drawable		d,
    unsigned long	valuemask,
    XGCValues*		values)
{
  GC gc;

  if ((gc = (GC)Xmalloc (sizeof(struct _XGC))) == NULL) {
    return (NULL);
  }
  gc->rects = 0;
  gc->dashes = 0;
  gc->ext_data = (XExtData *)Xmalloc (sizeof (_WGCExt));
  gc->values = initial_GC;
  gc->dirty = 0L;

  memcpy (gc->ext_data, &initial_Ext, sizeof (_WGCExt));

  valuemask &= (1L << (GCLastBit + 1)) - 1;
  if (valuemask) _XUpdateGCCache (gc, valuemask, values);

  _WUpdateGC (gc);

  gc->dirty = 0L;
  return gc;
}

int
XFreeGC (Display *display, GC gc)
{
  if (gc) XFree (gc);
  return 0;
}

int
_XUpdateGCCache (gc, mask, attr)
    register unsigned long mask;
    register XGCValues *attr;
    register GC gc;
    {
    register XGCValues *gv = &gc->values;

    if (mask & GCFunction)
        if (gv->function != attr->function) {
	  gv->function = attr->function;
	  gc->dirty |= GCFunction;
	}
	
    if (mask & GCPlaneMask)
        if (gv->plane_mask != attr->plane_mask) {
            gv->plane_mask = attr->plane_mask;
	    gc->dirty |= GCPlaneMask;
	  }

    if (mask & GCForeground)
        if (gv->foreground != attr->foreground) {
            gv->foreground = attr->foreground;
	    gc->dirty |= GCForeground;
	  }

    if (mask & GCBackground)
        if (gv->background != attr->background) {
            gv->background = attr->background;
	    gc->dirty |= GCBackground;
	  }

    if (mask & GCLineWidth)
        if (gv->line_width != attr->line_width) {
            gv->line_width = attr->line_width;
	    gc->dirty |= GCLineWidth;
	  }

    if (mask & GCLineStyle)
        if (gv->line_style != attr->line_style) {
            gv->line_style = attr->line_style;
	    gc->dirty |= GCLineStyle;
	  }

    if (mask & GCCapStyle)
        if (gv->cap_style != attr->cap_style) {
            gv->cap_style = attr->cap_style;
	    gc->dirty |= GCCapStyle;
	  }
    
    if (mask & GCJoinStyle)
        if (gv->join_style != attr->join_style) {
            gv->join_style = attr->join_style;
	    gc->dirty |= GCJoinStyle;
	  }

    if (mask & GCFillStyle)
        if (gv->fill_style != attr->fill_style) {
            gv->fill_style = attr->fill_style;
	    gc->dirty |= GCFillStyle;
	  }

    if (mask & GCFillRule)
        if (gv->fill_rule != attr->fill_rule) {
    	    gv->fill_rule = attr->fill_rule;
	    gc->dirty |= GCFillRule;
	  }

    if (mask & GCArcMode)
        if (gv->arc_mode != attr->arc_mode) {
	    gv->arc_mode = attr->arc_mode;
	    gc->dirty |= GCArcMode;
	  }

    /* always write through tile change, since client may have changed pixmap contents */
    if (mask & GCTile) {
	    gv->tile = attr->tile;
	    gc->dirty |= GCTile;
	  }

    /* always write through stipple change, since client may have changed pixmap contents */
    if (mask & GCStipple) {
	    gv->stipple = attr->stipple;
	    gc->dirty |= GCStipple;
	  }

    if (mask & GCTileStipXOrigin)
        if (gv->ts_x_origin != attr->ts_x_origin) {
    	    gv->ts_x_origin = attr->ts_x_origin;
	    gc->dirty |= GCTileStipXOrigin;
	  }

    if (mask & GCTileStipYOrigin)
        if (gv->ts_y_origin != attr->ts_y_origin) {
	    gv->ts_y_origin = attr->ts_y_origin;
	    gc->dirty |= GCTileStipYOrigin;
	  }

    if (mask & GCFont)
        if (gv->font != attr->font) {
	    gv->font = attr->font;
	    gc->dirty |= GCFont;
	  }

    if (mask & GCSubwindowMode)
        if (gv->subwindow_mode != attr->subwindow_mode) {
	    gv->subwindow_mode = attr->subwindow_mode;
	    gc->dirty |= GCSubwindowMode;
	  }

    if (mask & GCGraphicsExposures)
        if (gv->graphics_exposures != attr->graphics_exposures) {
	    gv->graphics_exposures = attr->graphics_exposures;
	    gc->dirty |= GCGraphicsExposures;
	  }

    if (mask & GCClipXOrigin)
        if (gv->clip_x_origin != attr->clip_x_origin) {
	    gv->clip_x_origin = attr->clip_x_origin;
	    gc->dirty |= GCClipXOrigin;
	  }

    if (mask & GCClipYOrigin)
        if (gv->clip_y_origin != attr->clip_y_origin) {
	    gv->clip_y_origin = attr->clip_y_origin;
	    gc->dirty |= GCClipYOrigin;
	  }

    /* always write through mask change, since client may have changed pixmap contents */
    if (mask & GCClipMask) {
	    gv->clip_mask = attr->clip_mask;
	    gc->dirty |= GCClipMask;
	    gc->rects = 0;
	  }

    if (mask & GCDashOffset)
        if (gv->dash_offset != attr->dash_offset) {
	    gv->dash_offset = attr->dash_offset;
	    gc->dirty |= GCDashOffset;
	  }

    if (mask & GCDashList)
        if ((gv->dashes != attr->dashes) || (gc->dashes == True)) {
            gv->dashes = attr->dashes;
	    gc->dirty |= GCDashList;
	    gc->dashes = 0;
	    }
    return 0;
    }

int
_XFlushGCCache(dpy, gc)
     Display *dpy;
     GC gc;
{
    if (gc->dirty) {
	gc->dirty = 0L; /* allow extensions to see dirty bits */
    }
    return 0;
}

GContext XGContextFromGC(
    GC			gc)
{
  return (gc->gid);
}

void XFlushGC(
    Display*		display,
    GC			gc)
{
  return;
}

int
XCopyGC (dpy, srcGC, mask, destGC)
     register Display *dpy;
     unsigned long mask;		/* which ones to set initially */
     GC srcGC, destGC;
{
    register XGCValues *destgv = &destGC->values,
    		       *srcgv = &srcGC->values;

    mask &= (1L << (GCLastBit + 1)) - 1;
    /* if some of the source values to be copied are "dirty", flush them 
       out before sending the CopyGC request. */
    if (srcGC->dirty & mask)
         _XFlushGCCache(dpy, srcGC);

    /* mark the copied values "not dirty" in the destination. */
    destGC->dirty &= ~mask;

    if (mask & GCFunction)
    	destgv->function = srcgv->function;
	
    if (mask & GCPlaneMask)
        destgv->plane_mask = srcgv->plane_mask;

    if (mask & GCForeground)
        destgv->foreground = srcgv->foreground;

    if (mask & GCBackground)
        destgv->background = srcgv->background;

    if (mask & GCLineWidth)
        destgv->line_width = srcgv->line_width;

    if (mask & GCLineStyle)
        destgv->line_style = srcgv->line_style;

    if (mask & GCCapStyle)
        destgv->cap_style = srcgv->cap_style;
    
    if (mask & GCJoinStyle)
        destgv->join_style = srcgv->join_style;

    if (mask & GCFillStyle)
    	destgv->fill_style = srcgv->fill_style;

    if (mask & GCFillRule) 
        destgv->fill_rule = srcgv->fill_rule;

    if (mask & GCArcMode)
        destgv->arc_mode = srcgv->arc_mode;

    if (mask & GCTile)
        destgv->tile = srcgv->tile;

    if (mask & GCStipple)
        destgv->stipple = srcgv->stipple;

    if (mask & GCTileStipXOrigin)
        destgv->ts_x_origin = srcgv->ts_x_origin;

    if (mask & GCTileStipYOrigin)
        destgv->ts_y_origin = srcgv->ts_y_origin;

    if (mask & GCFont) 
        destgv->font = srcgv->font;

    if (mask & GCSubwindowMode) 
        destgv->subwindow_mode = srcgv->subwindow_mode;

    if (mask & GCGraphicsExposures) 
        destgv->graphics_exposures = srcgv->graphics_exposures;

    if (mask & GCClipXOrigin) 
        destgv->clip_x_origin = srcgv->clip_x_origin;

    if (mask & GCClipYOrigin) 
        destgv->clip_y_origin = srcgv->clip_y_origin;

    if (mask & GCClipMask) {
	destGC->rects = srcGC->rects;
        destgv->clip_mask = srcgv->clip_mask;
	}

    if (mask & GCDashOffset) 
        destgv->dash_offset = srcgv->dash_offset;

    if (mask & GCDashList) {
	destGC->dashes = srcGC->dashes;
        destgv->dashes = srcgv->dashes;
	}
    return 0;
    }

int
XChangeGC (dpy, gc, valuemask, values)
    register Display *dpy;
    GC gc;
    unsigned long valuemask;
    XGCValues *values;
{
    valuemask &= (1L << (GCLastBit + 1)) - 1;
    if (valuemask) _XUpdateGCCache (gc, valuemask, values);

    _WUpdateGC (gc);

    /* if any Resource ID changed, must flush */
    if (gc->dirty & (GCFont | GCTile | GCStipple))
        _XFlushGCCache(dpy, gc);

    return 0;
}

/*
 * All gc fields except GCClipMask and GCDashList
 */
#define ValidGCValuesBits (GCFunction | GCPlaneMask | GCForeground | \
			   GCBackground | GCLineWidth | GCLineStyle | \
			   GCCapStyle | GCJoinStyle | GCFillStyle | \
			   GCFillRule | GCTile | GCStipple | \
			   GCTileStipXOrigin | GCTileStipYOrigin | \
			   GCFont | GCSubwindowMode | GCGraphicsExposures | \
			   GCClipXOrigin | GCClipYOrigin | GCDashOffset | \
			   GCArcMode)

/*ARGSUSED*/
Status XGetGCValues (dpy, gc, valuemask, values)
    Display *dpy;
    GC gc;
    unsigned long valuemask;
    XGCValues *values;
{
    if (valuemask == ValidGCValuesBits) {
	char dashes = values->dashes;
	Pixmap clip_mask = values->clip_mask;
	*values = gc->values;
	values->dashes = dashes;
	values->clip_mask = clip_mask;
	return True;
    }

    if (valuemask & ~ValidGCValuesBits) return False;

    if (valuemask & GCFunction)
      values->function = gc->values.function;

    if (valuemask & GCPlaneMask)
      values->plane_mask = gc->values.plane_mask;

    if (valuemask & GCForeground)
      values->foreground = gc->values.foreground;

    if (valuemask & GCBackground)
      values->background = gc->values.background;

    if (valuemask & GCLineWidth)
      values->line_width = gc->values.line_width;

    if (valuemask & GCLineStyle)
      values->line_style = gc->values.line_style;

    if (valuemask & GCCapStyle)
      values->cap_style = gc->values.cap_style;

    if (valuemask & GCJoinStyle)
      values->join_style = gc->values.join_style;

    if (valuemask & GCFillStyle)
      values->fill_style = gc->values.fill_style;

    if (valuemask & GCFillRule)
      values->fill_rule = gc->values.fill_rule;

    if (valuemask & GCTile)
      values->tile = gc->values.tile;

    if (valuemask & GCStipple)
      values->stipple = gc->values.stipple;

    if (valuemask & GCTileStipXOrigin)
      values->ts_x_origin = gc->values.ts_x_origin;

    if (valuemask & GCTileStipYOrigin)
      values->ts_y_origin = gc->values.ts_y_origin;

    if (valuemask & GCFont)
      values->font = gc->values.font;

    if (valuemask & GCSubwindowMode)
      values->subwindow_mode = gc->values.subwindow_mode;

    if (valuemask & GCGraphicsExposures)
      values->graphics_exposures = gc->values.graphics_exposures;

    if (valuemask & GCClipXOrigin)
      values->clip_x_origin = gc->values.clip_x_origin;

    if (valuemask & GCClipYOrigin)
      values->clip_y_origin = gc->values.clip_y_origin;

    if (valuemask & GCDashOffset)

      values->dash_offset = gc->values.dash_offset;

    if (valuemask & GCArcMode)
      values->arc_mode = gc->values.arc_mode;

    return True;
}

int
XSetForeground (Display *dpy, GC gc, unsigned long foreground)
{
  XGCValues values;
  values.foreground = foreground;
  return XChangeGC (dpy, gc, GCForeground, &values);
}

int
XSetBackground (Display *dpy, GC gc, unsigned long background)
{
  XGCValues values;
  values.background = background;
  return XChangeGC (dpy, gc, GCBackground, &values);
}

int
XSetFillStyle (Display *dpy, GC gc, int fill_style)
{
  XGCValues values;
  values.fill_style = fill_style;
  return XChangeGC (dpy, gc, GCFillStyle, &values);
}

int
XSetFont (Display *dpy, GC gc, Font font)
{
  XGCValues values;
  values.font = font;
  return XChangeGC (dpy, gc, GCFont, &values);
}

int
XSetLineAttributes(
    Display*		display,
    GC			gc,
    unsigned int	line_width,
    int			line_style,
    int			cap_style,
    int			join_style)
{
  XGCValues values;
  values.line_width = line_width;
  values.line_style = line_style;
  values.cap_style = cap_style;
  values.join_style = join_style;
  return XChangeGC (display, gc, GCLineWidth
		    | GCLineStyle
		    | GCCapStyle
		    | GCJoinStyle, &values);
}
