/* $Id: pixmap.c 1.2 1994/01/15 02:01:00 ulrich Exp $ */
/*
 * pixmap.c
 *
 * X library functions for pixmap creation.
 */
#include "Xlibemu.h"

static char swap_bits (char c)
{
  char s = 0;
  int i;

  for (i = 0; i < 8; i++) {
    s <<= 1;
    s |= (c & 1);
    c >>= 1;
  }
  return s;
}

Pixmap XCreatePixmap
(Display*	display,
 Drawable	d,
 unsigned int	width,
 unsigned int	height,
 unsigned int	depth)
{
  _WPixmap *pixmap;

  pixmap = (Pixmap) Xcalloc (1, sizeof (_WPixmap));
  if (pixmap != NULL) {
  }
  pixmap->type = 3;
  pixmap->context = GrCreateContext (width, height, NULL, NULL);

  return (Pixmap) pixmap;
}

Pixmap XCreateBitmapFromData(
    Display*		display,
    Drawable		d,
    _Xconst char*	data,
    unsigned int	width,
    unsigned int	height
)
{
  int i, pad, size;
  char *bmp_data;
  Pixmap bitmap;
  GrPattern *bmp;

  pad = (width + 7) >> 3;
  size = height * pad;
  bmp = (GrPattern *) Xmalloc (sizeof (GrPattern) + size);
  bmp_data = (char *) (bmp + 1);
  
  for (i = 0; i < height; i++)
    bmp_data[i] = swap_bits (data[i * pad]);

  bmp->gp_ispixmap = 0;
  bmp->gp_bmp_data = bmp_data;
  bmp->gp_bmp_height = height;
  bmp->gp_bmp_fgcolor = WhitePixel (display, DefaultScreen (display));
  bmp->gp_bmp_bgcolor = BlackPixel (display, DefaultScreen (display));
  bmp->gp_bitmap.bmp_memflags = 1;	/* GrDestroyPattern does not free bitmap data */

  bitmap = (_WPixmap *) Xcalloc (1, sizeof(bitmap[0]) + size);
  bitmap->type = 0;
  bitmap->pattern = bmp;
  bitmap->width = width;
  bitmap->height = height;
  bitmap->data = (char *) (bitmap + 1);
  memcpy (bitmap->data, data, size);

  return bitmap;
}

Pixmap XCreatePixmapFromBitmapData(
    Display*		display,
    Drawable		d,
    char*		data,
    unsigned int	width,
    unsigned int	height,
    unsigned long	fg,
    unsigned long	bg,
    unsigned int	depth)
{
  int i, pad, size;
  Pixmap pixmap;
  char *pixmap_data;

  pad = (width + 7) >> 3;
  size = pad * height;
  pixmap_data = (char *) alloca (size);
  for (i = 0; i < size; i++)
    pixmap_data[i] = swap_bits (data[i]);

  pixmap = (Pixmap) Xcalloc (1, sizeof (_WPixmap));
  pixmap->type = 1;
  pixmap->pattern = GrBuildPixmapFromBits (pixmap_data, width, height, fg, bg);
  pixmap->width = width;
  pixmap->height = height;
  return pixmap;
}

int XFreePixmap(
    Display*		display,
    Pixmap		pixmap)
{
  if (pixmap == None) return 0;
  _WDrawInvalid (pixmap);	/* flush drawable cache */
  if (pixmap->context) GrDestroyContext (pixmap->context);
  if (pixmap->pattern) GrDestroyPattern (pixmap->pattern);
  Xfree (pixmap);
  return 1;
}


