/* $Id: window.c 1.2 1994/01/15 02:04:34 ulrich Exp $ */
/*
 * window.c 
 *
 * Internal window tree functions for Xlibemu.
 */

#include "Xlibemu.h"

extern void _WContextFilledBox();

Bool
_WCheckWindow (Display *dpy, Window window, int request_code)
{
  if (window == None
      || (window->upper_sibling &&
	  window->upper_sibling->parent != window->parent)
      || (window->lower_sibling &&
	  window->lower_sibling->parent != window->parent)) {
    _WError (dpy, (XID) window, BadWindow, request_code);
    return 0;
  }
  return 1;
}

/* tree functions */

void
_WTreeDelete (Window w)
{
  if (w->upper_sibling)
    w->upper_sibling->lower_sibling = w->lower_sibling;
  else if (w->parent)
    w->parent->top_child = w->lower_sibling;

  if (w->lower_sibling)
    w->lower_sibling->upper_sibling = w->upper_sibling;
  else if (w->parent)
    w->parent->bottom_child = w->upper_sibling;

  w->parent = None;
  w->upper_sibling = None;
  w->lower_sibling = None;
}

void
_WTreeInsertTop (Window parent, Window window)
{
  if (parent == None) return;
  window->parent = parent;
  window->upper_sibling = None;
  window->lower_sibling = parent->top_child;
  if (parent->top_child)
    parent->top_child->upper_sibling = window;
  else
    parent->bottom_child = window;
  parent->top_child = window;
}

void
_WTreeInsertBot (Window parent, Window window)
{
  if (parent == None) return;
  window->parent = parent;
  window->lower_sibling = None;
  window->upper_sibling = parent->bottom_child;
  if (parent->bottom_child)
    parent->bottom_child->lower_sibling = window;
  else
    parent->top_child = window;
  parent->bottom_child = window;
}

void
_WTreeInsertAbove (Window sibling, Window window)
{
  window->parent = sibling->parent;
  window->lower_sibling = sibling;
  window->upper_sibling = sibling->upper_sibling;
  if (sibling->upper_sibling)
    sibling->upper_sibling->lower_sibling = window;
  else
    window->parent->top_child = window;
  sibling->upper_sibling = window;
}

void
_WTreeInsertBelow (Window sibling, Window window)
{
  window->parent = sibling->parent;
  window->lower_sibling = sibling->lower_sibling;
  window->upper_sibling = sibling;
  if (sibling->lower_sibling)
    sibling->lower_sibling->upper_sibling = window;
  else
    window->parent->bottom_child = window;
  sibling->lower_sibling = window;
}


/* Xlib calls for window handling */

Window
XCreateWindow
(   Display*		dpy,
    Window		parent,
    int			x,
    int			y,
    unsigned int	width,
    unsigned int	height,
    unsigned int	border_width,
    int			depth,
    unsigned int	class,
    Visual*		visual,
    unsigned long	valuemask,
    XSetWindowAttributes*	attributes)
{
  Window window;
  XEvent xe;
  
  if (dpy == None) return None;

  if (parent == None) {
    parent = DefaultRootWindow (dpy);
  }

  window = (Window) Xcalloc (1, sizeof (*window));
  if (window == None) {
    return None;
  }
  window->type		= 2;	/* 0 = bitmap, 1 = pixmap */
  window->parent	= parent;
  window->x		= x;
  window->y		= y;
  window->width		= width;
  window->height	= height;
  window->border_width	= border_width;
  window->depth		= (depth == CopyFromParent) ? parent->depth : depth;
  window->class		= (class == CopyFromParent) ? parent->class : class;
  window->visual	= (visual == CopyFromParent) ? parent->visual : visual;
  window->visible_region = XCreateRegion();

  _WTreeInsertTop (parent, window);
  /* Needs parent, x, y, width and height initialized */
  _WNewWindowContext (window);

  /* Set default functions for border and background */

  window->border.fillBox = _WDummyFilledBox;
  window->background.fillBox = _WDummyFilledBox;

  XChangeWindowAttributes (dpy, window, valuemask, attributes);

  if (window->parent &&
      window->parent->event_mask & SubstructureNotifyMask) {
    xe.xcreatewindow.type = CreateNotify;
    xe.xcreatewindow.send_event = 0;
    xe.xcreatewindow.display = dpy;
    xe.xcreatewindow.parent = parent;
    xe.xcreatewindow.window = window;
    xe.xcreatewindow.x = x;
    xe.xcreatewindow.y = y;
    xe.xcreatewindow.width = width;
    xe.xcreatewindow.height = height;
    xe.xcreatewindow.border_width = border_width;
    xe.xcreatewindow.override_redirect = window->override_redirect;
    _WDispatchEvent (&xe);
  }
  return window;
}


Window
XCreateSimpleWindow
(   Display*		dpy,
    Window		parent,
    int			x,
    int			y,
    unsigned int	width,
    unsigned int	height,
    unsigned int	border_width,
    unsigned long	border,
    unsigned long	background)
{
  int screen;
  XSetWindowAttributes xswa;

  if (dpy == None) {
    return None;
  }

  screen = DefaultScreen (dpy);
  xswa.border_pixel = border;
  xswa.background_pixel = background;

  return XCreateWindow (dpy, parent, x, y, width, height, border_width,
			DisplayPlanes (dpy, screen),
			InputOutput,
			DefaultVisual (dpy, screen),
			CWBorderPixel | CWBackPixel, &xswa);
}

int
XChangeWindowAttributes
(   Display*		dpy,
    Window		window,
    unsigned long	valuemask,
    XSetWindowAttributes* attributes)
{
  int redraw_border;

  if (0 == _WCheckWindow (dpy, window, X_ChangeWindowAttributes))
    return 0;

  redraw_border = False;

  if (valuemask & CWBackPixmap) {
    window->background.fillArg = (long) None;
    window->background.fillBox = _WDummyFilledBox;
    if (attributes->background_pixmap != None) {
      if (attributes->background_pixmap->pattern != NULL) {
	window->background.fillArg = (long) attributes->background_pixmap->pattern;
	window->background.fillBox = GrPatternFilledBox;
      }
      else if (attributes->background_pixmap->context != NULL) {
	window->background.fillArg = (long) attributes->background_pixmap->context;
	window->background.fillBox = _WContextFilledBox;
      }
    }
  }
  if (valuemask & CWBackPixel) {
    window->background.fillArg	= attributes->background_pixel;
    window->background.fillBox	= GrFilledBox;
  }
  if (valuemask & CWBorderPixmap) {
    window->border.fillArg	= (long) attributes->border_pixmap->pattern;
    window->border.fillBox	= (window->border.fillArg != None)
      ? GrPatternFilledBox : _WDummyFilledBox;
    redraw_border = True;
  }
  if (valuemask & CWBorderPixel) {
    window->border.fillArg	= attributes->border_pixel;
    window->border.fillBox	= GrFilledBox;
    redraw_border = True;
  }
  if (valuemask & CWBitGravity)
    window->bit_gravity		= attributes->bit_gravity;
  if (valuemask & CWWinGravity)
    window->win_gravity		= attributes->win_gravity;
  if (valuemask & CWBackingStore)
    window->backing_store	= attributes->backing_store;
  if (valuemask & CWBackingPlanes)
    window->backing_planes	= attributes->backing_planes;
  if (valuemask & CWBackingPixel)
    window->backing_pixel	= attributes->backing_pixel;
  if (valuemask & CWSaveUnder)
    window->save_under		= attributes->save_under;
  if (valuemask & CWEventMask) {
    int i;
    long mask;

    window->event_masks[ConnectionNumber(dpy)] = attributes->event_mask;
    mask = 0;
    for (i = 0; i < MAX_CONN; i++) {
      mask |= window->event_masks[i];
    }
    window->event_mask = mask;
  }
  if (valuemask & CWDontPropagate)
    window->do_not_propagate_mask	= attributes->do_not_propagate_mask;
  if (valuemask & CWOverrideRedirect)
    window->override_redirect	= attributes->override_redirect;
  if (valuemask & CWColormap)
    window->colormap		= attributes->colormap;
  if (valuemask & CWCursor)
    window->cursor		= attributes->cursor;

  if (window->viewable) {
    if (redraw_border)
      _WDrawWindowBorder (dpy, window, &window->border_port, True);
  }
  
  return 1;
}

Status XGetWindowAttributes(
     register Display *dpy,
     Window w,
     XWindowAttributes *attr)
{
  attr->x = w->x;
  attr->y = w->y;
  attr->width = w->width;
  attr->height = w->height;
  attr->border_width = w->border_width;
  attr->depth = w->depth;
  attr->visual = w->visual;
  attr->root = DefaultRootWindow (dpy);
  attr->class = w->class;
  attr->bit_gravity = w->bit_gravity;
  attr->win_gravity = w->win_gravity;
  attr->backing_store = NotUseful /* w->backing_store */;
  attr->backing_planes = 0 /* w->backing_planes */;
  attr->backing_pixel = 0 /* w->backing_pixel */;
  attr->save_under = w->save_under;
  attr->colormap = w->colormap;
  attr->map_installed = True;
  attr->map_state = (w->mapped) ? IsViewable : IsUnmapped;
  attr->all_event_masks = w->event_mask;
  attr->your_event_mask = w->event_masks[ConnectionNumber(dpy)];
  attr->do_not_propagate_mask = w->do_not_propagate_mask;
  attr->override_redirect = w->override_redirect;
  attr->screen = &dpy->screens[0];
  return 1;
}


int
XSelectInput(
    Display*		dpy,
    Window		window,
    long		event_mask)
{
  XSetWindowAttributes attr;

  attr.event_mask = event_mask;
  return XChangeWindowAttributes (dpy, window, CWEventMask, &attr);
}


int
XMapRaised (Display* dpy, Window window)
{
  XRaiseWindow (dpy, window);
  return XMapWindow (dpy, window);
}

void
_WMapWindow (Display* dpy, Window window)
{
  int viewable;

  window->mapped = 1;
  viewable = (window->parent)
    ? window->parent->viewable
    : True;
  if (viewable != window->viewable) {
    window->viewable = viewable;
    /* set viewable for all subwindows depending on map state */
    _WSetViewable (window, viewable);
  }
  if (window->class == InputOutput) {
    if (window->parent) {
      _WSetVisibleSubwindowsInBox (window->parent, &window->border_port);
      _WInvalidateParent (window);
    }
    else {
      _WSetVisibleRegion (window);
      _WSetVisibleSubwindows (window);
    }
    _WDrawWindow (dpy, window, &window->border_port, True);
    _WDrawSubwindows (dpy, window, &window->border_port, True);
  }
}

int
XMapWindow (Display* dpy, Window window)
{
  XEvent xe;

  if (0 == _WCheckWindow (dpy, window, X_MapWindow))
    return 0;

  if (window->mapped)
    return 1;

  if (window->override_redirect == False
      && window->parent != NULL
      && window->parent->event_mask & SubstructureRedirectMask) {
    xe.xmaprequest.type = MapRequest;
    xe.xmaprequest.send_event = False;
    xe.xmaprequest.parent = window->parent;
    xe.xmaprequest.window = window;
    if (_WRedirectEvent (dpy, &xe))
      return 0;
  }

  _WMapWindow (dpy, window);

  if (window->event_mask & StructureNotifyMask) {
    xe.xmap.type = MapNotify;
    xe.xmap.send_event = 0;
    xe.xmap.display = dpy;
    xe.xmap.event = window;
    xe.xmap.window = window;
    xe.xmap.override_redirect = window->override_redirect;
    _WDispatchEvent (&xe);
  }
  if (window->parent &&
      window->parent->event_mask & SubstructureNotifyMask) {
    xe.xmap.type = MapNotify;
    xe.xmap.send_event= 0;
    xe.xmap.display = dpy;
    xe.xmap.event = window->parent;
    xe.xmap.window = window;
    xe.xmap.override_redirect = window->override_redirect;
    _WDispatchEvent (&xe);
  }
  return 1;
}


int
XMapSubwindows (Display* dpy, Window window)
{
  if (0 == _WCheckWindow (dpy, window, X_MapSubwindows))
    return 0;

  _WSetMapping (window, True);
  if (window->viewable) {
    _WSetViewable (window, True);
    _WSetVisibleRegion (window);
    _WSetVisibleSubwindows (window);
    _WDrawSubwindows (dpy, window, &window->window_port, True);
  }
  return 1;
}

int
XUnmapSubwindows (Display* dpy, Window window)
{
  Window sibling;
  Region region;

  if (0 == _WCheckWindow (dpy, window, X_UnmapSubwindows))
    return 0;
  _WSetMapping (window, False);
  _WSetViewable (window, False);
  _WSetVisibleRegion (window);

#if 0
  if (window->viewable) {
    region = XCreateRegion ();
    for (sibling = window->top_child;
	 sibling != None;
	 sibling = sibling->lower_sibling) {
      REGION temp;
      temp.rects = &temp.extents;
      temp.numRects = 1;
      temp.extents = sibling->border_port;
      temp.size = 1;
      XUnionRegion (&temp, region, region);
    }
    if (REGION_NOT_EMPTY(region))
      _WDrawWindowBackgroundRegion (dpy, window, region, True);
  }
#else
  _WDrawWindow (dpy, window, &window->window_port, True);
#endif
  return 1;
}

void
_WUnmapWindow (Display* dpy, Window window)
{
  window->mapped = 0;
  window->viewable = 0;
  _WSetViewable (window, 0);	/* make all subwindows unviewable */
  if (window->parent != None &&
      window->class == InputOutput) {
    _WSetVisibleRegion (window->parent);
    _WSetVisibleSubwindowsInBox (window->parent, &window->border_port);
    _WDrawSiblingsBelow (dpy, window, True);
    _WDrawWindowBackground (dpy, window->parent, &window->border_port, True);
  }
}

int
XUnmapWindow (Display* dpy, Window window)
{
  XEvent xe;

  if (0 == _WCheckWindow (dpy, window, X_UnmapWindow))
    return 0;
  if (! window->mapped)
    return 1;

  _WUnmapWindow (dpy, window);

  if (window->event_mask & StructureNotifyMask) {
    xe.xunmap.type = UnmapNotify;
    xe.xunmap.send_event = False;
    xe.xunmap.display = dpy;
    xe.xunmap.event = window;
    xe.xunmap.window = window;
    xe.xunmap.from_configure = False;
    _WDispatchEvent (&xe);
  }
  if (window->parent &&
      window->parent->event_mask & SubstructureNotifyMask) {
    xe.xunmap.type = UnmapNotify;
    xe.xunmap.send_event = False;
    xe.xunmap.display = dpy;
    xe.xunmap.event = window->parent;
    xe.xunmap.window = window;
    xe.xunmap.from_configure = False;
    _WDispatchEvent(&xe);
  }
  return 1;
}

Status XWithdrawWindow(
    Display*		display,
    Window		w,
    int			screen_number)
{
  XEvent xe;

  XUnmapWindow (display, w);

  xe.xunmap.type = UnmapNotify;
  xe.xunmap.send_event = False;
  xe.xunmap.display = display;
  xe.xunmap.event = RootWindow (display, screen_number);
  xe.xunmap.window = w;
  xe.xunmap.from_configure = False;
  _WDispatchEvent (&xe);
  return 1;
}

int
XDefineCursor
(Display*	dpy,
 Window		window,
 Cursor		cursor)
{
  XSetWindowAttributes attr;

  attr.cursor = cursor;
  return XChangeWindowAttributes (dpy, window, CWCursor, &attr);
}

int
XUndefineCursor
(Display*	dpy,
 Window		window)
{
  XSetWindowAttributes attr;

  attr.cursor = None;
  return XChangeWindowAttributes (dpy, window, CWCursor, &attr);
}

int
XMoveResizeWindow (Display *dpy, Window window, int x, int y,
		   unsigned int width, unsigned int height)
{
  XWindowChanges xwc;
  unsigned long values;

  xwc.x = x;
  xwc.y = y;
  xwc.width = width;
  xwc.height = height;
  values = CWX | CWY | CWWidth | CWHeight;
  return XConfigureWindow (dpy, window, values, &xwc);
}

int
XMoveWindow (Display *dpy, Window window, int x, int y)
{
  XWindowChanges xwc;
  unsigned long values;

  xwc.x = x;
  xwc.y = y;
  values = CWX | CWY;
  return XConfigureWindow (dpy, window, values, &xwc);
}

int
XResizeWindow (Display *dpy, Window window, unsigned int width, unsigned int height)
{
  XWindowChanges xwc;
  unsigned long values;

  xwc.width = width;
  xwc.height = height;
  values = CWWidth | CWHeight;
  return XConfigureWindow (dpy, window, values, &xwc);
}


int
XSetWindowBorder (Display *dpy, Window window, unsigned long pixel)
{
  XSetWindowAttributes attr;
  
  attr.border_pixel = pixel;
  return XChangeWindowAttributes (dpy, window, CWBorderPixel, &attr);
}

int
XSetWindowBackground (Display *dpy, Window window, unsigned long pixel)
{
  XSetWindowAttributes attr;

  attr.background_pixel = pixel;
  return XChangeWindowAttributes (dpy, window, CWBackPixel, &attr);
}

int
XSetWindowBorderWidth (Display *dpy, Window window, unsigned int border_width)
{
  XWindowChanges xwc;
  unsigned long values;

  xwc.border_width = border_width;
  values = CWBorderWidth;
  return XConfigureWindow (dpy, window, values, &xwc);
}

int
XSetWindowBorderPixmap (Display *dpy, Window window, Pixmap pixmap)
{
  XSetWindowAttributes attr;

  attr.border_pixmap = pixmap;
  return XChangeWindowAttributes (dpy, window, CWBorderPixmap, &attr);
}

int
XSetWindowBackgroundPixmap (Display *dpy, Window window, Pixmap pixmap)
{
  XSetWindowAttributes attr;

  attr.background_pixmap = pixmap;
  return XChangeWindowAttributes (dpy, window, CWBackPixmap, &attr);
}

int
XCirculateSubwindows (Display *dpy, Window parent, int direction)
{
  if (0 == _WCheckWindow (dpy, parent, X_CirculateWindow))
    return 0;
  switch (direction) {
  case RaiseLowest:
    if (parent->bottom_child != parent->top_child)
      XRaiseWindow (dpy, parent->bottom_child);
    break;
  case LowerHighest:
    if (parent->bottom_child != parent->top_child)
      XLowerWindow (dpy, parent->top_child);
    break;
  }
  return 0;
}

int
XCirculateSubwindowsUp (Display *dpy, Window parent)
{
  return XCirculateSubwindows (dpy, parent, LowerHighest);
}

int
XCirculateSubwindowsDown (Display *dpy, Window parent)
{
  return XCirculateSubwindows (dpy, parent, RaiseLowest);
}

int
XRaiseWindow (Display *dpy, Window window)
{
  if (0 == _WCheckWindow (dpy, window, X_ConfigureWindow))
    return 0;
  if (window->upper_sibling) {
    Window parent, sibling;

    parent = window->parent;
    _WTreeDelete (window);
    _WTreeInsertTop (parent, window);
    
    if (window->mapped && window->class == InputOutput) {
      for (sibling = window->lower_sibling;
	   sibling != None;
	   sibling = sibling->lower_sibling) {
	if (! sibling->mapped) continue;
	if (! EXTENTCHECK (&sibling->border_port,
			   &window->border_port)) continue;
	_WSetVisibleRegion (sibling);
	_WSetVisibleSubwindows (sibling);
      }
      _WSetVisibleRegion (window);
      _WSetVisibleSubwindows (window);
      _WDrawWindow (dpy, window, &window->border_port, True);
      _WDrawSubwindows (dpy, window, &window->border_port, True);
    }
  }
  return 1;
}

int
XLowerWindow (Display *dpy, Window window)
{
  if (0 == _WCheckWindow (dpy, window, X_ConfigureWindow))
    return 0;
  if (window == DefaultRootWindow (dpy))
    return 0;

  if (window->lower_sibling) {
    Window sibling, parent;

    parent = window->parent;
    _WTreeDelete (window);
    _WTreeInsertBot (parent, window);

    if (window->mapped && window->class == InputOutput) {
      for (sibling = window->upper_sibling;
	   sibling != None;
	   sibling = sibling->upper_sibling) {
	if (! sibling->mapped) continue;
	if (! EXTENTCHECK (&sibling->border_port,
			   &window->border_port)) continue;
	_WSetVisibleRegion (sibling);
	_WSetVisibleSubwindows (sibling);
	_WDrawWindow (dpy, sibling, &window->border_port, True);
	_WDrawSubwindows (dpy, sibling, &window->border_port, True);
      }
      _WSetVisibleRegion (window);
      _WSetVisibleSubwindows (window);
    }
  }
  return 1;
}

int
XReparentWindow (Display *dpy, Window window, Window parent, int x, int y)
{
  int mapped;
  XEvent xe;

  if (0 == _WCheckWindow (dpy, window, X_ReparentWindow))
    return 0;
  if (window == DefaultRootWindow (dpy) ||
      window->override_redirect != False)
    return 0;
  if (parent == NULL)
    parent = DefaultRootWindow (dpy);
  if (parent == window->parent)
    return 0;

  if (window->event_mask & StructureNotifyMask) {
    xe.xreparent.type = ReparentNotify;
    xe.xreparent.send_event = False;
    xe.xreparent.event = window;
    xe.xreparent.window = window;
    xe.xreparent.parent = parent;
    xe.xreparent.x = x;
    xe.xreparent.y = y;
    xe.xreparent.override_redirect = window->override_redirect;
    _WDispatchEvent (&xe);
  }
  if (window->parent != None &&
      window->parent->event_mask & SubstructureNotifyMask) {
    xe.xreparent.type = ReparentNotify;
    xe.xreparent.send_event = False;
    xe.xreparent.event = window->parent;
    xe.xreparent.window = window;
    xe.xreparent.parent = parent;
    xe.xreparent.x = x;
    xe.xreparent.y = y;
    xe.xreparent.override_redirect = window->override_redirect;
    _WDispatchEvent (&xe);
  }
  mapped = window->mapped;
  if (mapped) _WUnmapWindow (dpy, window);
  _WTreeDelete (window);
  _WTreeInsertTop (parent, window);
  /* move window */
  window->x = x;
  window->y = y;
  _WNewWindowContext (window);
  if (mapped) _WMapWindow (dpy, window);
  return 1;
}

void
_WDestroyWindow (Display *dpy, Window w)
{
  if (w->mapped) XUnmapWindow (dpy, w);

  _WDrawInvalid (w);	/* flush drawable cache */
  _WTreeDelete (w);

  w->event_mask = 0;
  w->do_not_propagate_mask = ~0;
  w->visibility = VisibilityFullyObscured;
  EMPTY_REGION(w->visible_region);

  GrDestroyContext (w->context);
#if 0
  /* the event queue may contain references to w */

  XDestroyRegion (w->visible_region);
  Xfree (w);
#endif
}

int
XDestroyWindow (Display *dpy, Window window)
{
  if (0 == _WCheckWindow (dpy, window, X_DestroyWindow))
    return 0;
  XDestroySubwindows (dpy, window);
  _WDestroyWindow (dpy, window);
  return 1;
}

int
XDestroySubwindows (Display *dpy, Window parent)
{
  Window window;

  if (0 == _WCheckWindow (dpy, parent, X_DestroySubwindows))
    return 0;
  for (window = parent->top_child;
       window != NULL;
       window = window->lower_sibling) {
    XDestroySubwindows (dpy, window);
    _WDestroyWindow (dpy, window);
  }
  return 1;
}

int
XClearWindow
(Display*	dpy,
 Window		window)
{
  return XClearArea (dpy, window, 0, 0, 0, 0, False);
}

int
XClearArea
(Display*	dpy,
 Window		window, 
 int		x,
 int		y,
 unsigned int	width,
 unsigned int	height,
 Bool		exposures)
{
  if (0 == _WCheckWindow (dpy, window, X_ClearArea))
    return 0;
  if (width == 0) width = window->width;
  if (height == 0) height = window->height;

  if (! _WDrawContext ((Drawable) window, None))
    return 0;
  _WDrawRegion (window->background.fillBox,
		x, y, x + width - 1, y + height - 1,
		window->background.fillArg);

  /*
   * for the future draw region should only expose visible
   * rextangles.
   */
  if (exposures) {
    XEvent xe;
    xe.xexpose.type = Expose;
    xe.xexpose.send_event = 0;
    xe.xexpose.display = dpy;
    xe.xexpose.window = window;
    xe.xexpose.x = x;
    xe.xexpose.y = y;
    xe.xexpose.width = width;
    xe.xexpose.height = height;
    xe.xexpose.count = 0;
    _WDispatchEvent (&xe);
  }

  return 1;
}

Status
XGetGeometry(
    Display*		dpy,
    Drawable		d,
    Window*		root_return,
    int*		x_return,
    int*		y_return,
    unsigned int*	width_return,
    unsigned int*	height_return,
    unsigned int*	border_width_return,
    unsigned int*	depth_return)
{
  Window w;

  switch (d->type) {
  case 2:
    w = (Window) d;
    if (0 == _WCheckWindow (dpy, w, X_GetGeometry))
      return 0;
    *root_return = DefaultRootWindow (dpy);
    *x_return = w->x;
    *y_return = w->y;
    *width_return = w->width;
    *height_return = w->height;
    *border_width_return = w->border_width;
    *depth_return = w->depth;
    return 1;
  case 0:
  case 1:
  case 3:
    *root_return = DefaultRootWindow (dpy);
    *x_return = 0;
    *y_return = 0;
    *width_return = ((Pixmap) d)->width;
    *height_return = ((Pixmap) d)->height;
    *border_width_return = 0;
    *depth_return = (d->type == 0) ? 1 : DefaultDepth (dpy, DefaultScreen (dpy));
    return 1;
  }
  return 0;
}

Bool XTranslateCoordinates(
    Display*		dpy,
    Window		src_w,
    Window		dest_w,
    int			src_x,
    int			src_y,
    int*		dest_x_return,
    int*		dest_y_return,
    Window*		child_return)
{
  int root_x, root_y;
  Window child;

  if (0 == _WCheckWindow (dpy, src_w, X_TranslateCoords))
    return 0;
  if (0 == _WCheckWindow (dpy, dest_w, X_TranslateCoords))
    return 0;

  root_x = (int) src_w->window_port.x1 + src_x;
  root_y = (int) src_w->window_port.y1 + src_y;
  child = _WPointToWindow (dpy, root_x, root_y);
  for (; child != None; child = child->parent) {
    if (child->parent == dest_w) break;
  }
  *dest_x_return = root_x - (int) dest_w->window_port.x1;
  *dest_y_return = root_y - (int) dest_w->window_port.y1;
  *child_return  = child;
  return True;
}

int XConfigureWindow(
    Display*		dpy,
    Window		w,
    unsigned int	value_mask,
    XWindowChanges*	values)		 
{
  int mapped;
  int i, x, y, width, height;
  XEvent xe;
  Window sibling, window;

  if (0 == _WCheckWindow (dpy, w, X_ConfigureWindow))
    return 0;
  i = ConnectionNumber(dpy);

  if (w->override_redirect == False
      && w->parent != None
      && 0 != (w->parent->event_mask & SubstructureRedirectMask)
      && 0 == (w->parent->event_masks[i] & SubstructureRedirectMask)) {
    xe.xconfigurerequest.type = ConfigureRequest;
    xe.xconfigurerequest.send_event = False;
    xe.xconfigurerequest.parent = w->parent;
    xe.xconfigurerequest.window = w;
    xe.xconfigurerequest.x = (value_mask & CWX) ? values->x : w->x;
    xe.xconfigurerequest.y = (value_mask & CWY) ? values->y : w->y;
    xe.xconfigurerequest.width = (value_mask & CWWidth) ? values->width : w->width;
    xe.xconfigurerequest.height = (value_mask & CWHeight) ? values->height : w->height;
    xe.xconfigurerequest.border_width = 
      (value_mask & CWBorderWidth) ? values->border_width : w->border_width;
    xe.xconfigurerequest.above = (value_mask & CWSibling) ? Above : None;
    xe.xconfigurerequest.detail = (value_mask & CWStackMode) ? values->stack_mode : 0;
    xe.xconfigurerequest.value_mask = value_mask;
    if (_WRedirectEvent (dpy, &xe))
      return 0;
  }
  if (w->override_redirect == False
      && 0 != (w->event_mask & ResizeRedirectMask)
      && 0 == (w->event_masks[i] & ResizeRedirectMask)
      && 0 != (value_mask & (CWWidth | CWHeight))) {
    xe.xresizerequest.type = ResizeRequest;
    xe.xresizerequest.send_event = False;
    xe.xresizerequest.window = w;
    xe.xresizerequest.width = (value_mask & CWWidth) ? values->width : w->width;
    xe.xresizerequest.height = (value_mask & CWHeight) ? values->height : w->height;
    if (_WRedirectEvent (dpy, &xe))
      return 0;
  }

  mapped = w->mapped;

  x = w->x;
  y = w->y;
  width = w->width;
  height = w->height;
  if (value_mask & CWX) {
    if (x != values->x)
      x = values->x;
    else
      value_mask &= ~CWX;
  }
  if (value_mask & CWY) {
    if (y != values->y)
      y = values->y;
    else
      value_mask &= ~CWY;
  }
  if (value_mask & CWWidth) {
    if (width != values->width)
      width = values->width;
    else
      value_mask &= ~CWWidth;
  }
  if (value_mask & CWHeight) {
    if (height != values->height)
      height = values->height;
    else
      value_mask &= ~CWHeight;
  }
  if (value_mask & (CWX | CWY | CWWidth | CWHeight)) {
    if (w->mapped) _WUnmapWindow (dpy, w);
    w->x = x;
    w->y = y;
    w->width = width;
    w->height = height;
    _WNewWindowContext (w);
  }

  if (value_mask & CWBorderWidth) {
    if (w->border_width != values->border_width) {
      if (w->mapped) _WUnmapWindow (dpy, w);
      w->border_width = values->border_width;
      _WGetBorderPort (w, &w->border_port);
    }
  }

  if (value_mask & CWStackMode) {

    if (w->mapped) _WUnmapWindow (dpy, w);

    if (value_mask & CWSibling) {
      sibling = values->sibling;
      if (sibling->parent == w->parent) {
	switch (values->stack_mode) {
	case Above:
	  _WTreeDelete (w);
	  _WTreeInsertAbove (sibling, w);
	  break;
	case Below:
	  _WTreeDelete (w);
	  _WTreeInsertBelow (sibling, w);
	  break;
	case TopIf:
	  for (window = w;
	       window != None;
	       window = window->upper_sibling) {
	    if (window == sibling
		&& EXTENTCHECK(&sibling->border_port, &w->window_port)) {
	      _WTreeDelete (w);
	      _WTreeInsertAbove (sibling, w);
	    }
	  }	    
	  break;
	case BottomIf:
	  for (window = w;
	       window != None;
	       window = window->lower_sibling) {
	    if (window == sibling
		&& EXTENTCHECK(&sibling->border_port, &w->window_port)) {
	      _WTreeDelete (w);
	      _WTreeInsertBelow (sibling, w);
	    }
	  }	    
	  break;
	case Opposite:
	  break;
	}
      }
    }
    else {
      switch (values->stack_mode) {
      case Above:
	XRaiseWindow (dpy, w);
	break;
      case Below:
	XLowerWindow (dpy, w);
	break;
      case TopIf:
	for (sibling = w->upper_sibling;
	     sibling != None;
	     sibling = sibling->upper_sibling) {
	  if (EXTENTCHECK (&w->window_port, &sibling->border_port)) {
	    XRaiseWindow (dpy, w);
	    break;
	  }
	}
	break;
      case BottomIf:
	for (sibling = w->lower_sibling;
	     sibling != None;
	     sibling = sibling->upper_sibling) {
	  if (EXTENTCHECK (&w->border_port, &sibling->window_port)) {
	    XLowerWindow (dpy, w);
	    break;
	  }
	}
	break;
      case Opposite:
	break;
      }
    }
  }

  if (w->event_mask & StructureNotifyMask) {
    xe.xconfigure.type = ConfigureNotify;
    xe.xconfigure.send_event = 0;
    xe.xconfigure.display = dpy;
    xe.xconfigure.event = w;
    xe.xconfigure.window = w;
    xe.xconfigure.x = w->x;
    xe.xconfigure.y = w->y;
    xe.xconfigure.width = w->width;
    xe.xconfigure.height = w->height;
    xe.xconfigure.border_width = w->border_width;
    xe.xconfigure.above = w->lower_sibling;
    xe.xconfigure.override_redirect = w->override_redirect;
    _WDispatchEvent (&xe);
  }
  if (w->parent != None &&
      w->parent->event_mask & SubstructureNotifyMask) {
    xe.xconfigure.type = ConfigureNotify;
    xe.xconfigure.send_event = 0;
    xe.xconfigure.display = dpy;
    xe.xconfigure.event = w->parent;
    xe.xconfigure.window = w;
    xe.xconfigure.x = w->x;
    xe.xconfigure.y = w->y;
    xe.xconfigure.width = w->width;
    xe.xconfigure.height = w->height;
    xe.xconfigure.border_width = w->border_width;
    xe.xconfigure.above = w->lower_sibling;
    xe.xconfigure.override_redirect = w->override_redirect;
    _WDispatchEvent (&xe);
  }
  if (mapped != 0 && w->mapped == 0) _WMapWindow (dpy, w);
  return 0;
}

int
XQueryTree
(Display *	dpy,
 Window		w,
 Window *	root_return,
 Window *	parent_return,
 Window **	children_return,
 unsigned int * nchildren_return)
{
  Window child, *children;
  int n;

  if (0 == _WCheckWindow (dpy, w, X_QueryTree))
    return 0;
  n = 0;
  for (child = w->bottom_child;
       child != None;
       child = child->lower_sibling) {
    n++;
  }
  children = (Window *) Xmalloc ((n+1) * sizeof (Window));
  n = 0;
  for (child = w->bottom_child;
       child != None;
       child = child->lower_sibling) {
    children[n++] = child;
  }
  *root_return = DefaultRootWindow (dpy);
  *parent_return = w->parent;
  *children_return = children;
  *nchildren_return = n;
  return 1;
}

int XSetWindowColormap(
    Display*		display,
    Window		w,
    Colormap		colormap)
{
  w->colormap = colormap;
  return 0;
}
